"""Test fitness assignment processes."""


from math import inf, isfinite
from typing import Final, cast

from numpy.random import Generator, default_rng
from pycommons.types import type_error

from moptipy.algorithms.so.fitness import Fitness, FRecord, check_fitness
from moptipy.api.objective import Objective
from moptipy.api.operators import Op0
from moptipy.api.space import Space
from moptipy.tests.component import validate_component
from moptipy.utils.nputils import rand_seed_generate


class _FRecord(FRecord):
    """The internal F-record."""

    def __init__(self, x, z):
        """Initialize."""
        super().__init__(x, inf)
        #: the internal id
        self.z: Final[int] = z


def validate_fitness(fitness: Fitness, objective: Objective, space: Space,
                     op0: Op0) -> None:
    """
    Validate a fitness assignment process on a given problem.

    :param fitness: the fitness assignment process
    :param objective: the objective function
    :param space: the space of solutions
    :param op0: the nullary operator
    """
    if not isinstance(fitness, Fitness):
        raise type_error(fitness, "fitness", Fitness)
    check_fitness(fitness)
    validate_component(fitness)

    random: Final[Generator] = default_rng()
    pop1: Final[list[_FRecord]] = []
    pop2: Final[list[_FRecord]] = []
    pop3: Final[list[_FRecord]] = []
    ps: Final[int] = int(1 + random.integers(48))
    for i in range(ps):
        fr: _FRecord = _FRecord(space.create(), i)
        op0.op0(random, fr.x)
        fr.f = objective.evaluate(fr.x)
        fr.it = int(random.integers(1, 20))
        if fr.fitness != inf:
            raise ValueError(f"v = {fr.fitness}, should be inf")
        pop1.append(fr)
        fr2: _FRecord = _FRecord(fr.x, fr.z)
        fr2.f = fr.f
        fr2.it = fr.it
        pop2.append(fr2)
        fr2 = _FRecord(fr.x, fr.z)
        fr2.f = fr.f
        fr2.it = fr.it
        pop3.append(fr2)

    for k in range(6):
        if k >= 3:
            # make all records identical
            fr0 = pop1[0]
            for i in range(ps):
                fr = _FRecord(fr0.x, i)
                fr.f = fr0.f
                fr.it = fr0.it
                fr.fitness = fr0.fitness
                pop1[i] = fr
                fr = _FRecord(fr0.x, i)
                fr.f = fr0.f
                fr.it = fr0.it
                fr.fitness = fr0.fitness
                pop2[i] = fr
                fr = _FRecord(fr0.x, i)
                fr.f = fr0.f
                fr.it = fr0.it
                fr.fitness = fr0.fitness
                pop3[i] = fr
        if k in (2, 4):
            for fr in pop1:
                if random.integers(2) <= 0:
                    fr.fitness = inf if random.integers(2) <= 0 else -inf

        seed: int = rand_seed_generate()
        fitness.initialize()
        fitness.assign_fitness(cast(list[FRecord], pop1), default_rng(seed))
        fitness.initialize()
        fitness.assign_fitness(cast(list[FRecord], pop3), default_rng(seed))
        pop1.sort(key=lambda r: r.z)
        pop3.sort(key=lambda r: r.z)

        for i, fr in enumerate(pop1):
            if not isinstance(fr, _FRecord):
                raise type_error(fr, f"pop[{i}]", _FRecord)
            fr2 = pop2[i]
            if fr.x is not fr2.x:
                raise ValueError("fitness assignment changed x reference!")
            if fr.it is not fr2.it:
                raise ValueError(
                    f"fitness assignment assigned rec.it to {fr.it}!")
            if fr.f is not fr2.f:
                raise ValueError(
                    f"fitness assignment assigned rec.f to {fr.f}!")
            if not isinstance(fr.fitness, int | float):
                raise type_error(fr.fitness, "rec.fitness", (int, float))
            if not isfinite(fr.fitness):
                raise ValueError(
                    f"rec.fitness should be finite, but is {fr.fitness}")
            fr2 = pop3[i]
            if (fr2.fitness != fr.fitness) or (fr2.f is not fr.f) or \
                    (fr2.it is not fr.it) or (fr2.x is not fr.x):
                raise ValueError(f"inconsistency detected when repeating "
                                 f"fitness assignment: {str(fr2)!r} != "
                                 f"{str(fr)!r} at index {i} of population "
                                 f"of length {len(pop1)}.")



"""
A binary operator copying each bit from either source string.

Uniform crossover copies the value of every bit with the same probability
from either of the two parent strings. This means that all bits where both
parent strings have the same value remain unchanged and are copied directly
to the offspring. The bits where the parent strings have different values
are effectively randomized. This is easy to see from the fact that, if both
parents have different values for one bit, then one of them must have the
bit set to `1` and the other one set to `0`. This means that the value in the
offspring is set to `0` with probability `0.5` and set to `1` with probability
`0.5`. This corresponds to drawing it uniformly at random, i.e., randomizing
it.

1. Gilbert Syswerda. Uniform Crossover in Genetic Algorithms. In J. David
   Schaffer, editor, *Proceedings of the 3rd International Conference on
   Genetic Algorithms* (ICGA'89), June 4-7, 1989, Fairfax, VA, USA, pages 2-9.
   San Francisco, CA, USA: Morgan Kaufmann Publishers Inc.
   ISBN: 1-55860-066-3. https://www.researchgate.net/publication/201976488.
2. Hans-Georg Beyer and Hans-Paul Schwefel. Evolution Strategies - A
   Comprehensive Introduction. *Natural Computing: An International
   Journal* 1(1):3-52, March 2002, http://doi.org/10.1023/A:1015059928466.
   https://www.researchgate.net/publication/220132816.
3. Hans-Georg Beyer. An Alternative Explanation for the Manner in which
   Genetic Algorithms Operate. *Biosystems* 41(1):1-15, January 1997,
   https://doi.org/10.1016/S0303-2647(96)01657-7.
4. William M. Spears and Kenneth Alan De Jong. On the Virtues of Parameterized
   Uniform Crossover. In Richard K. Belew and Lashon Bernard Booker, editors,
   *Proceedings of the Fourth International Conference on Genetic Algorithms*
   (ICGA'91), July 13-16, 1991, San Diego, CA, USA, pages 230-236.
   San Francisco, CA, USA: Morgan Kaufmann Publishers Inc.
   ISBN: 1-55860-208-9. https://www.mli.gmu.edu/papers/91-95/91-18.pdf.
"""

import numba  # type: ignore
import numpy as np
from numpy.random import Generator

from moptipy.api.operators import Op2


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def uniform(random: Generator, dest: np.ndarray, x0: np.ndarray,
            x1: np.ndarray) -> None:
    """
    Perform the actual work of the uniform crossover.

    :param random: the random number generator
    :param dest: the destination array
    :param x0: the first source array
    :param x1: the second source array

    >>> a = np.full(10, True)
    >>> b = np.full(len(a), False)
    >>> r = np.random.default_rng(10)
    >>> out = np.empty(len(a), bool)
    >>> uniform(r, out, a, b)
    >>> print(out)
    [ True  True False False  True  True  True False  True  True]
    >>> uniform(r, out, a, b)
    >>> print(out)
    [False False False  True False  True False False  True  True]
    >>> uniform(r, out, a, b)
    >>> print(out)
    [False  True False False  True  True  True  True  True  True]
    >>> uniform(r, out, a, b)
    >>> print(out)
    [False  True  True False  True  True False False False  True]
    """
    for i in range(len(dest)):  # pylint: disable=C0200
        v = random.integers(0, 2)  # create boolean value
        # copy from x0 with p=0.5 and from x1 with p=0.5
        dest[i] = x1[i] if v == 0 else x0[i]


class Op2Uniform(Op2):
    """
    A binary search operation that copies each bit from either source.

    For each index `i` in the destination array `dest`, uniform
    crossover copies the value from the first source string `x0`with
    probability 0.5 and otherwise the value from the second source
    string `x1`. All bits that have the same value in `x0` and `x1`
    will retain this value in `dest`, all bits where `x0` and `x1`
    differ will effectively be randomized (be `0` with probability 0.5
    and `1` with probability 0.5).
    """

    def __init__(self):
        """Initialize the uniform crossover operator."""
        super().__init__()
        self.op2 = uniform  # type: ignore

    def __str__(self) -> str:
        """
        Get the name of this binary operator.

        :return: "uniform"
        """
        return "uniform"



"""
The operators for :mod:`moptipy.spaces.ordered_choices` spaces.

- Module :mod:`~moptipy.operators.ordered_choices.op0_choose_and_shuffle`
  provides a nullary operator for randomly picking an ordered choices
  instance.
"""



"""A nullary operator filling a bit string with random values."""

import numpy as np
from numpy.random import Generator

from moptipy.api.operators import Op0


class Op0Random(Op0):
    """Fill a bit string with random values."""

    def op0(self, random: Generator, dest: np.ndarray) -> None:
        """
        Fill the string `dest` with random values.

        :param random: the random number generator
        :param dest: the bit string to fill. Afterwards, its contents will
            be random.
        """
        np.copyto(dest, random.integers(0, 2, dest.shape, dest.dtype))

    def __str__(self) -> str:
        """
        Get the name of this operator.

        :return: "randomize"
        """
        return "randomize"



"""Utilities for creating and storing figures."""
import os.path
import statistics as st
import warnings
from math import isfinite, sqrt
from typing import Any, Callable, Final, Iterable, Sequence, cast

import matplotlib.pyplot as plt  # type: ignore
from matplotlib import rcParams  # type: ignore
from matplotlib.artist import Artist  # type: ignore
from matplotlib.axes import Axes  # type: ignore
from matplotlib.backend_bases import RendererBase  # type: ignore
from matplotlib.backends.backend_agg import RendererAgg  # type: ignore
from matplotlib.figure import Figure, SubFigure  # type: ignore
from pycommons.io.path import Path, directory_path
from pycommons.types import check_int_range, type_error, type_name_of

import moptipy.utils.plot_defaults as pd
from moptipy.utils.lang import Lang

# Ensure that matplotlib uses Type 1 fonts.
# Some scientific conferences, such as GECCO organized by ACM, require this.
# In the language utilities :meth:`~moptipy.utils.lang.Lang.font`, we do
# return acceptable fonts anyway, but it may be better to set this here
# explicitly to avoid any problem.
rcParams["pdf.fonttype"] = 42
rcParams["ps.fonttype"] = 42

#: The golden ratio constant with value 1.618033988749895.
__GOLDEN_RATIO: Final[float] = 0.5 + (0.5 * sqrt(5))


def create_figure(width: float | int | None = 8.6,
                  height: float | int | None = None,
                  dpi: float | int | None = 384.0,
                  **kwargs) -> Figure:
    """
    Create a matplotlib figure.

    :param width: the optional width
    :param height: the optional height
    :param dpi: the dpi value
    :param kwargs: a set of optional arguments
    :return: the figure option
    """
    # Check the size, i.e., width and height and use
    put_size: bool = True
    if width is None:
        if height is None:
            put_size = False
        else:
            width = __GOLDEN_RATIO * height
    elif height is None:
        height = width / __GOLDEN_RATIO

    if put_size:
        if isinstance(height, int):
            height = float(height)
        if not isinstance(height, float):
            raise ValueError(f"Invalid height type {type(height)}.")
        if not (isfinite(height) and (0.1 < height < 10000.0)):
            raise ValueError(f"Invalid height {height}.")
        nheight = int(0.5 + (height * 72)) / 72.0
        if not (isfinite(nheight) and (0.1 < nheight < 10000.0)):
            raise ValueError(f"Invalid height {height} as it maps to "
                             f"{nheight} after to-point-and-back conversion.")

        if isinstance(width, int):
            width = float(width)
        if not isinstance(width, float):
            raise ValueError(f"Invalid width type {type(width)}.")
        if not (isfinite(width) and (0.1 < width < 10000.0)):
            raise ValueError(f"Invalid width {width}.")
        nwidth = int(0.5 + (width * 72)) / 72.0
        if not (isfinite(nwidth) and (0.1 < nwidth < 10000.0)):
            raise ValueError(f"Invalid width {width} as it maps to "
                             f"{nwidth} after to-point-and-back conversion.")

        kwargs["figsize"] = width, height

    if dpi is not None:
        if isinstance(dpi, int):
            dpi = float(dpi)
        if not isinstance(dpi, float):
            raise ValueError(f"Invalid dpi type {type(dpi)}.")
        if not (isfinite(dpi) and (1.0 < dpi < 10000.0)):
            raise ValueError(f"Invalid dpi value {dpi}.")
        kwargs["dpi"] = dpi

    if "frameon" not in kwargs:
        kwargs["frameon"] = False

    if "constrained_layout" not in kwargs:
        kwargs["constrained_layout"] = False

    return Figure(**kwargs)


def __divide_evenly(items: int, chunks: int, reverse: bool) -> list[int]:
    """
    Divide `n` items into `k` chunks, trying to create equally-sized chunks.

    :param items: the number of items to divide
    :param chunks: the number of chunks
    :param reverse: should we put the additional items at the end (True)
        or front (False)
    :returns: the list of items

    >>> print(__divide_evenly(9, 3, reverse=True))
    [3, 3, 3]
    >>> print(__divide_evenly(10, 3, reverse=True))
    [3, 3, 4]
    >>> print(__divide_evenly(10, 3, reverse=False))
    [4, 3, 3]
    >>> print(__divide_evenly(11, 3, reverse=True))
    [3, 4, 4]
    >>> print(__divide_evenly(11, 3, reverse=False))
    [4, 4, 3]
    >>> print(__divide_evenly(12, 3, reverse=False))
    [4, 4, 4]
    """
    # validate that we do not have more chunks than items or otherwise invalid
    # parameters
    if (items <= 0) or (chunks <= 0) or (chunks > items):
        raise ValueError(f"cannot divide {items} items into {chunks} chunks.")

    # First we compute the minimum number of items per chunk.
    # Our basic solution is to put exactly this many items into each chunk.
    # For example, if items=10 and chunks=3, this will fill 3 items into each
    # chunk.
    result: list[int] = [items // chunks] * chunks
    # We then fill the remaining items into the chunks at the front
    # (reverse=False) or end (reverse=True) of the list.
    # We do this by putting exactly one such item into each chunk, starting
    # from the end.
    # Since items modulo chunks must be in [0..chunks), this is possible.
    # This setup will yield the lowest possible standard deviation of items
    # per chunk.
    # In case of items=10 and chunks=3, items % chunks = 1 and the chunks will
    # become [3, 3, 4]. If items=3, no items need to be distributed and we get
    # [3, 3, 3].
    if reverse:
        for i in range(1, (items % chunks) + 1):
            result[-i] += 1
    else:
        for i in range(items % chunks):
            result[i] += 1
    return result


def create_figure_with_subplots(
        items: int,
        max_items_per_plot: int = 3,
        max_rows: int = 3,
        max_cols: int = 1,
        min_rows: int = 1,
        min_cols: int = 1,
        default_width_per_col: float | int | None = 8.6,
        max_width: float | int | None = 8.6,
        default_height_per_row: float | int | None = 8.6 / __GOLDEN_RATIO,
        max_height: float | int | None = 9,
        dpi: float | int | None = 384.0,
        plot_config: dict[str, Any] | None = None,
        **kwargs) \
        -> tuple[Figure, tuple[tuple[Axes | Figure,
                                     int, int, int, int, int], ...]]:
    """
    Divide a figure into nrows*ncols sub-plots.

    :param items: the number of items to divide
    :param max_items_per_plot: the maximum number of items per plot
    :param max_rows: the maximum number of rows
    :param max_cols: the maximum number of columns
    :param min_rows: the minimum number of rows
    :param min_cols: the minimum number of cols
    :param default_width_per_col: the optional default width of a column
    :param default_height_per_row: the optional default height per row
    :param max_height: the maximum height
    :param max_width: the maximum width
    :param dpi: the dpi value
    :param kwargs: a set of optional arguments
    :param plot_config: the configuration to be applied to all sub-plots
    :returns: a tuple with the figure, followed by a series of tuples with
        each sub-figure, the index of the first item assigned to it, the
        index of the last item assigned to it + 1, their row, their column,
        and their overall index
    """
    # First, we do a lot of sanity checks
    check_int_range(items, "items", 1, 1_000_000)
    check_int_range(max_items_per_plot, "max_items_per_plot", 1, 1_000_000)
    check_int_range(max_rows, "max_rows", 1, 100)
    check_int_range(min_rows, "min_rows", 1, max_rows)
    check_int_range(max_cols, "max_cols", 1, 100)
    check_int_range(min_cols, "min_cols", 1, max_cols)
    if (max_cols * max_rows * max_items_per_plot) < items:
        raise ValueError(
            f"Cannot distribute {items} items into at most {max_rows} rows "
            f"and {max_cols} cols with at most {max_items_per_plot} per "
            "plot.")
    if default_width_per_col is not None:
        default_width_per_col = float(default_width_per_col)
        if (not isfinite(default_width_per_col)) \
                or (default_width_per_col <= 0.1) \
                or (default_width_per_col >= 1000):
            raise ValueError(
                f"invalid default_width_per_col {default_width_per_col}")
    if max_width is not None:
        max_width = float(max_width)
        if (not isfinite(max_width)) \
                or (max_width <= 0.1) or (max_width >= 10000):
            raise ValueError(f"invalid max_width {max_width}")
        if (default_width_per_col is not None) \
                and (default_width_per_col > max_width):
            raise ValueError(f"default_width_per_col {default_width_per_col} "
                             f"> max_width {max_width}")
    if default_height_per_row is not None:
        default_height_per_row = float(default_height_per_row)
        if (not isfinite(default_height_per_row)) \
                or (default_height_per_row <= 0.1) \
                or (default_height_per_row >= 10000):
            raise ValueError(
                f"invalid default_height_per_row {default_height_per_row}")
    if max_height is not None:
        max_height = float(max_height)
        if (not isfinite(max_height)) \
                or (max_height <= 0.1) or (max_height >= 10000):
            raise ValueError(f"invalid max_width {max_height}")
        if (default_height_per_row is not None) \
                and (max_height < default_height_per_row):
            raise ValueError(
                f"max_height {max_height} < "
                f"default_height_per_row {default_height_per_row}")

    # setting up the default dimensions
    if default_height_per_row is None:
        default_height_per_row = 1.95
        if max_height is not None:
            default_height_per_row = min(default_height_per_row, max_height)
    if max_height is None:
        max_height = max_rows * default_height_per_row
    if default_width_per_col is None:
        default_width_per_col = 8.6
        if max_width is not None:
            default_width_per_col = min(default_width_per_col, max_width)
    if max_width is None:
        max_width = default_width_per_col
    if (plot_config is not None) and not isinstance(plot_config, dict):
        raise type_error(plot_config, "plot_config", (dict, None))

    # How many plots do we need at least? Well, if we can put
    # max_items_per_plot items into each plot and have items many items, then
    # we need ceil(items / max_items_per_plot) many. This is what we compute
    # here.
    min_plots: Final[int] = (items // max_items_per_plot) \
        + min(1, items % max_items_per_plot)
    # The maximum conceivable number of plots would be items, so this we do
    # not need to compute.
    plots_i: Final[int] = 6
    rows_i: Final[int] = plots_i + 1
    cols_i: Final[int] = rows_i + 1
    height_i: Final[int] = cols_i + 1
    width_i: Final[int] = height_i + 1
    plots_per_row_i: Final[int] = width_i + 1
    chunks_i: Final[int] = plots_per_row_i + 1
    best: tuple[float,  # the plots-per-row std
                float,  # the items-per-plot std
                float,  # the overall area of the figure
                float,  # the std deviation of (w, h*golden_ratio),
                float,  # -plot area
                float,  # the std deviation of (w, h*golden_ratio) per plot
                int,  # the number of plots
                int,  # the number of rows
                int,  # the number of cols
                float,  # the figure height
                float,  # the figure width
                list[int],  # the plots per row
                list[int]] = \
        (1000000.0, 1000000.0, 1000000.0, 1000000.0, 1000000.0, 1000000.0,
         1 << 62, 1 << 62, 1 << 62, 1000000.0, 1000000.0, [0], [0])

    # Now we simply try all valid combinations.
    for rows in range(min_rows, max_rows + 1):
        if rows > items:
            continue  # more rows than items
        for cols in range(min_cols, max_cols + 1):
            # We need cols plots per row, except for the last row, where we
            # could only put a single plot if we wanted.
            min_plots_for_config: int = 1 + ((rows - 1) * cols)
            if min_plots_for_config > items:
                continue  # if this exceeds the maximum number of plots, skip
            # compute the maximum plots we can get in this config
            max_plots_per_config: int = rows * cols
            if max_plots_per_config < min_plots:
                continue  # not enough plots possible

            # now we can iterate over the number of plots we can provide
            for plots in range(max(min_plots_for_config, min_plots),
                               min(max_plots_per_config, items) + 1):
                # We compute the figure width. If we have multiple columns,
                # we try to add 2% plus 0.02" as breathing space per column.
                fig_width = min(max_width, ((1.02 * default_width_per_col)
                                            + 0.02) * cols)
                # The width of a single plot is then computed as follows:
                plot_width = fig_width / cols
                if plot_width < 0.9:
                    continue  # the single plots would be too small
                # We compute the overall figure height. Again we add breathing
                # space if we have multiple rows.
                fig_height = min(max_height, ((1.02 * default_height_per_row)
                                              + 0.02) * rows)
                plot_height = fig_height / rows
                if plot_height < 0.9:
                    continue  # the single plots would be too small
                # So dimension-wise, the plot sizes are OK.
                # How about the distribution of items? We put the rows with
                # plots at the top and the plots with more items at the end.
                plot_distr = __divide_evenly(plots, rows, reverse=False)
                item_distr = __divide_evenly(items, plots, reverse=True)

                current: tuple[float, float, float, float, float, float,
                               int, int, int, float, float,
                               list[int], list[int]] = (
                    st.stdev(plot_distr) if rows > 1 else 0,
                    st.stdev(item_distr) if plots > 1 else 0,
                    fig_height * fig_width,  # the area of the figure
                    st.stdev([fig_width, fig_height * __GOLDEN_RATIO]),
                    -plot_width * plot_height,  # -plot area
                    st.stdev([plot_width, plot_height * __GOLDEN_RATIO]),
                    plots,  # the number of plots
                    rows,  # the number of rows
                    cols,  # the number of cols
                    fig_height,  # the figure height
                    fig_width,  # the figure width
                    plot_distr,  # the plots per row
                    item_distr)  # the data chunks per plot
                if ((current[plots_i] + 1)  # type: ignore
                    < best[plots_i]) or ((  # type: ignore
                        current[plots_i]   # type: ignore
                        <= (best[plots_i] + 1))  # type: ignore
                        and (current < best)):  # type: ignore
                    best = current

    n_plots: Final[int] = best[plots_i]  # type: ignore
    if n_plots > items:
        raise ValueError(
            f"Could not place {items} in {min_cols}..{max_cols} columns "
            f"at {min_rows}..{max_rows} rows with at most "
            f"{max_items_per_plot} items per plot for a max_width={max_width}"
            f" and max_height={max_height}.")

    # create the figure of the computed dimensions
    figure: Final[Figure] = create_figure(
        width=best[width_i], height=best[height_i],  # type: ignore
        dpi=dpi, **kwargs)
    if (n_plots <= 1) and (plot_config is None):
        # if there is only one plot, we are done here
        return figure, ((figure, 0, items, 0, 0, 0), )

    # if there are multiple plots, we need to generate them
    allfigs: list[tuple[Axes | Figure,
                        int, int, int, int, int]] = []
    index: int = 0
    chunk_start: int = 0
    nrows: Final[int] = best[rows_i]  # type: ignore
    ncols: Final[int] = best[cols_i]  # type: ignore
    chunks: Final[list[int]] = best[chunks_i]  # type: ignore
    plots_per_row: Final[list[int]] = best[plots_per_row_i]  # type: ignore
    for i in range(nrows):
        for j in range(plots_per_row[i]):
            chunk_next = chunk_start + chunks[index]
            if plot_config is None:
                sp = figure.add_subplot(nrows, ncols, (i * ncols) + j + 1)
            else:
                sp = figure.add_subplot(nrows, ncols, (i * ncols) + j + 1,
                                        **plot_config)
            allfigs.append((sp, chunk_start, chunk_next, i, j, index))
            chunk_start = chunk_next
            index += 1
    return figure, tuple(allfigs)


def save_figure(fig: Figure,
                file_name: str = "figure",
                dir_name: str = ".",
                formats: str | Iterable[str] = "svg") -> list[Path]:
    """
    Store the given figure in files of the given formats and dispose it.

    :param fig: the figure to save
    :param file_name: the basic file name
    :param dir_name: the directory name
    :param formats: the file format(s)
    :return: a list of files
    """
    if not isinstance(fig, Figure):
        raise type_error(fig, "figure", Figure)
    if not isinstance(file_name, str):
        raise type_error(file_name, "file_name", str)
    if len(file_name) <= 0:
        raise ValueError(f"Invalid filename {file_name!r}.")
    if not isinstance(dir_name, str):
        raise type_error(dir_name, "dir_name", str)
    if len(dir_name) <= 0:
        raise ValueError(f"Invalid dirname {dir_name!r}.")
    if isinstance(formats, str):
        formats = [formats]
    if not isinstance(formats, Iterable):
        raise type_error(formats, "formats", Iterable)

    size = fig.get_size_inches()
    orientation: Final[str] = \
        "landscape" if size[0] >= size[1] else "portrait"

    # set minimal margins to the axes to avoid wasting space
    for ax in fig.axes:  # consider the dimension of the axes
        margins: list[int] = [0] * len(ax.margins())
        ax.margins(*margins)

    use_dir = directory_path(dir_name)
    files = []
    for fmt in formats:
        if not isinstance(fmt, str):
            raise type_error(fmt, "element of formats", str)
        dest_file = Path(os.path.join(use_dir, f"{file_name}.{fmt}"))
        dest_file.ensure_file_exists()
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            # UserWarning: There are no gridspecs with layoutgrids.
            # Possibly did not call parent GridSpec with the "figure" keyword
            fig.savefig(dest_file, transparent=True, format=fmt,
                        orientation=orientation,
                        dpi="figure",
                        bbox_inches="tight",
                        pad_inches=1.0 / 72.0)
        dest_file.enforce_file()
        files.append(dest_file)

    fig.clf(False)
    plt.close(fig)
    del fig

    if len(files) <= 0:
        raise ValueError("No formats were specified.")

    return files


def label_box(axes: Axes,
              text: str,
              x: float | None = None,
              y: float | None = None,
              font_size: float = pd.importance_to_font_size(0),
              may_rotate_text: bool = False,
              z_order: float | None = None,
              font: None | str | Callable =
              lambda: Lang.current().font()) -> None:
    """
    Put a label text box near an axis.

    :param axes: the axes to add the label to
    :param text: the text to place
    :param x: the location along the x-axis: `0` means left,
        `0.5` means centered, `1` means right
    :param y: the location along the x-axis: `0` means bottom,
        `0.5` means centered, `1` means top
    :param font_size: the font size
    :param may_rotate_text: should we rotate the text by 90° if that
        makes sense (`True`) or always keep it horizontally (`False`)
    :param z_order: an optional z-order value
    :param font: the font to use
    """
    if x is None:
        if y is None:
            raise ValueError("At least one of x or y must not be None.")
        x = 0
    elif y is None:
        y = 0

    spacing: Final[float] = max(4.0, font_size / 2.0)
    xtext: float = 0.0
    ytext: float = 0.0
    xalign: str = "center"
    yalign: str = "center"
    if x >= 0.85:
        xtext = -spacing
        xalign = "right"
    elif x <= 0.15:
        xtext = spacing
        xalign = "left"
    if y >= 0.85:
        ytext = -spacing
        yalign = "top"
    elif y <= 0.15:
        ytext = spacing
        yalign = "bottom"

    args = {"text": text,
            "xy": (x, y),
            "xytext": (xtext, ytext),
            "verticalalignment": yalign,
            "horizontalalignment": xalign,
            "xycoords": "axes fraction",
            "textcoords": "offset points",
            "fontsize": font_size,
            "bbox": {"boxstyle": "round",
                     "color": "white",
                     "fill": True,
                     "linewidth": 0,
                     "alpha": 0.9}}
    if z_order is not None:
        args["zorder"] = z_order

    if may_rotate_text and (len(text) > 2):
        args["rotation"] = 90

    if callable(font):
        font = font()
    if font is not None:
        if not isinstance(font, str):
            raise type_error(font, "font", str)
        args["fontname"] = font

    axes.annotate(**args)  # type: ignore


def label_axes(axes: Axes,
               x_label: str | None = None,
               x_label_inside: bool = True,
               x_label_location: float = 0.5,
               y_label: str | None = None,
               y_label_inside: bool = True,
               y_label_location: float = 1,
               font_size: float = pd.importance_to_font_size(0),
               z_order: float | None = None) -> None:
    """
    Put labels on a figure.

    :param axes: the axes to add the label to
    :param x_label: a callable returning the label for
        the x-axis, a label string, or `None` if no label should be put
    :param x_label_inside: put the x-axis label inside the plot (so that
        it does not consume additional vertical space)
    :param x_label_location: the location of the x-axis label if it is
        placed inside the plot area
    :param y_label: a callable returning the label for
        the y-axis, a label string, or `None` if no label should be put
    :param y_label_inside: put the xyaxis label inside the plot (so that
        it does not consume additional horizontal space)nal vertical space)
    :param y_label_location: the location of the y-axis label if it is
        placed inside the plot area
    :param font_size: the font size to use
    :param z_order: an optional z-order value
    """
    # put the label on the x-axis, if any
    if x_label is not None:
        if not isinstance(x_label, str):
            raise type_error(x_label, "x_label", str)
        if len(x_label) > 0:
            if x_label_inside:
                label_box(axes, text=x_label, x=x_label_location, y=0,
                          font_size=font_size, z_order=z_order)
            else:
                axes.set_xlabel(x_label, fontsize=font_size)

    # put the label on the y-axis, if any
    if y_label is not None:
        if not isinstance(y_label, str):
            raise type_error(y_label, "y_label", str)
        if len(y_label) > 0:
            if y_label_inside:
                label_box(axes, text=y_label, x=0, y=y_label_location,
                          font_size=font_size, may_rotate_text=True,
                          z_order=z_order)
            else:
                axes.set_ylabel(y_label, fontsize=font_size)


def get_axes(figure: Axes | Figure) -> Axes:
    """
    Obtain the axes from a figure or axes object.

    :param figure: the figure
    :return: the Axes
    """
    if isinstance(figure, Figure):
        return figure.add_axes((0.005, 0.005, 0.99, 0.99))
    if hasattr(figure, "axes") \
            or isinstance(getattr(type(figure), "axes", None), property):
        try:
            if isinstance(figure.axes, Axes):
                return cast(Axes, figure.axes)
            if isinstance(figure.axes, Sequence):
                ax = figure.axes[0]
                if isinstance(ax, Axes):
                    return cast(Axes, ax)
            elif isinstance(figure.axes, Iterable):
                for k in figure.axes:
                    if isinstance(k, Axes):
                        return cast(Axes, k)
                    break
        except TypeError:
            pass
        except IndexError:
            pass
    if isinstance(figure, Axes):
        return cast(Axes, figure)
    raise TypeError(
        f"Cannot get Axes of object of type {type_name_of(figure)}.")


def get_renderer(figure: Axes | Figure | SubFigure) -> RendererBase:
    """
    Get a renderer that can be used for determining figure element sizes.

    :param figure: the figure element
    :return: the renderer
    """
    if isinstance(figure, Axes):
        figure = figure.figure
    if not isinstance(figure, Figure):
        raise type_error(figure, "figure", Figure)
    canvas = figure.canvas
    if hasattr(canvas, "renderer"):
        return canvas.renderer
    if hasattr(canvas, "get_renderer"):
        return canvas.get_renderer()
    return RendererAgg(width=figure.get_figwidth(),
                       height=figure.get_figheight(),
                       dpi=figure.get_dpi())


def cm_to_inch(cm: int | float) -> float:
    """
    Convert cm to inch.

    :param cm: the cm value
    :return: the value in inch
    """
    if not isinstance(cm, int):
        if not isinstance(cm, float):
            raise type_error(cm, "cm", (int, float))
        if not isfinite(cm):
            raise ValueError(f"cm must be finite, but is {cm}.")
    res: float = cm / 2.54
    if not isfinite(res):
        raise ValueError(f"Conversation {cm} cm to inch "
                         f"must be finite, but is {res}.")
    return res


#: the color attributes
__COLOR_ATTRS: Final[tuple[tuple[bool, bool, str], ...]] = \
    ((True, True, "get_label"), (False, True, "label"),
     (False, True, "_label"), (True, False, "get_color"),
     (False, False, "color"), (False, False, "_color"),
     (False, False, "edgecolor"), (False, False, "_edgecolor"),
     (False, False, "markeredgecolor"), (False, False, "_markeredgecolor"))


def get_label_colors(
        handles: Iterable[Artist],
        color_map: dict[str, tuple[float, ...] | str] | None = None,
        default_color: tuple[float, ...] | str = pd.COLOR_BLACK) \
        -> list[tuple[float, ...] | str]:
    """
    Get a list with label colors from a set of artists.

    :param handles: the handles
    :param color_map: an optional color decode
    :param default_color: the default color
    :returns: a list of label colors
    """
    if not isinstance(handles, Iterable):
        raise type_error(handles, "handles", Iterable)

    def __get_color(a: Artist, colmap=color_map,
                    defcol=default_color) -> tuple[float, ...] | str:
        if not isinstance(a, Artist):
            raise type_error(a, "artist", Artist)

        for acall, astr, aname in __COLOR_ATTRS:
            if hasattr(a, aname):
                val = getattr(a, aname)
                if val is None:
                    continue
                if acall:
                    val = val()
                if val is None:
                    continue
                if astr:
                    if colmap:
                        if val in colmap:
                            val = colmap[val]
                        else:
                            continue
                    else:
                        continue
                if val is None:
                    continue
                if val != defcol:
                    return val
        return defcol

    return [__get_color(aa) for aa in handles]



"""
In this package we provide operators for bit strings.

- Module :mod:`~moptipy.operators.bitstrings.op0_random` offers a nullary
  search operator that samples bit strings of a fixed length in a uniform
  random fashion.
- Module :mod:`~moptipy.operators.bitstrings.op1_flip1` offers a unary
  operator that flips exactly one bit.
- Module :mod:`~moptipy.operators.bitstrings.op1_flip_m` offers a unary
  operator with step size (see
  :class:`~moptipy.api.operators.Op1WithStepSize`). A step size of `0.0`
  means that it will flip one bit, a step size of `1.0` means that it flips
  all bits.
- Module :mod:`~moptipy.operators.bitstrings.op1_m_over_n_flip` flips each
  bit with probability `m/n` but ensures that at least one bit is flipped.
  This results in a number of bits being flipped being drawn from a Binomial
  distribution (and re-drawn if `0` was drawn).
- Module :mod:`~moptipy.operators.bitstrings.op2_uniform` offers the binary
  uniform crossover operation which copies each bit with probability `0.5`
  from the first and with the same probability from the second parent.
"""



"""
Provide the functionality to access search and solution spaces.

A :class:`Space` is the abstraction of the data structures for solutions and
points in the search space that need to be generated, copied, and stored
during the optimization process. This allows us to develop black-box
algorithms while still being able to properly remember the best solutions,
storing them as text strings in log files, and to validate whether they are
correct.

All search or solution spaces in `moptipy` inherit from :class:`Space`. If
you implement a new space, you should test it with the pre-defined unit test
routine :func:`~moptipy.tests.space.validate_space`.

The following pre-defined spaces are currently available:

- :class:`~moptipy.spaces.bitstrings.BitStrings`, the space of `n`-dimensional
  bit strings
- :class:`~moptipy.spaces.intspace.IntSpace`, a space of `n`-dimensional
  integer strings, where each element is between predefined inclusive bounds
  `min_value...max_value`.
- :class:`~moptipy.spaces.permutations.Permutations` is a special version of
  the :class:`~moptipy.spaces.intspace.IntSpace` where all elements are
  permutations of a base string
  :attr:`~moptipy.spaces.permutations.Permutations.blueprint`. This means that
  it can represent permutations both with and without repetitions. Depending
  on the base string, each element may occur an element-specific number of
  times. For the base string `(-1, -1, 2, 7, 7, 7)`, for example, `-1` may
  occur twice, `2` can occur once, and `7` three times.
- :class:`~moptipy.spaces.vectorspace.VectorSpace` is the space of
  `n`-dimensional floating point number vectors.
"""

from typing import Any

from pycommons.types import type_error

from moptipy.api.component import Component


# start book
class Space(Component):
    """
    A class to represent both search and solution spaces.

    The space basically defines a container data structure and basic
    operations that we can apply to them. For example, a solution
    space contains all the possible solutions to an optimization
    problem. All of them are instances of one data structure. An
    optimization as well as a black-box process needs to be able to
    create and copy such objects. In order to store the solutions we
    found in a text file, we must further be able to translate them to
    strings. We should also be able to parse such strings. It is also
    important to detect whether two objects are the same and whether
    the contents of an object are valid. All of this functionality is
    offered by the `Space` class.
    """

    def create(self) -> Any:
        # end book
        """
        Generate an instance of the data structure managed by the space.

        The state/contents of this data structure are undefined. It may
        not pass the :meth:`validate` method.

        :return: the new instance
        """

    def copy(self, dest, source) -> None:  # +book
        """
        Copy one instance of the data structure to another one.

        Notice that the first parameter of this method is the destination,
        which will be overwritten by copying the contents of the second
        parameter over it.

        :param dest: the destination data structure,
            whose contents will be overwritten with those from `source`
        :param source: the source data structure, which remains
            unchanged and whose contents will be copied to `dest`
        """

    def to_str(self, x) -> str:  # +book
        """
        Obtain a textual representation of an instance of the data structure.

        This method should convert an element of the space to a string
        representation that is parseable by :meth:from_str: and should ideally
        not be too verbose. For example, when converting a list or array `x`
        of integers to a string, one could simply do
        `";".join([str(xx) for xx in x])`, which would convert it to a
        semicolon-separated list without any wasted space.

        Notice that this method is used by the
        :class:`~moptipy.utils.logger.Logger` when storing the final
        optimization results in the log files in form of a
        :class:`~moptipy.utils.logger.TextLogSection` created via
        :meth:`~moptipy.utils.logger.Logger.text`. By implementing
        :meth:from_str: and :meth:to_str: as exact inverse of each other, you
        can thus ensure that you can always automatically load the results of
        your optimization runs from the log files created via
        :meth:`~moptipy.api.execution.Execution.set_log_file` of the
        :class:`~moptipy.api.execution.Execution` class.

        :param x: the instance
        :return: the string representation of x
        """

    def from_str(self, text: str) -> Any:  # +book
        """
        Transform a string `text` to one element of the space.

        This method should be implemented as inverse to :meth:to_str:.
        It should check the validity of the result before returning it.
        It may not always be possible to implement this method, but you
        should try.

        :param text: the input string
        :return: the element in the space corresponding to `text`
        """

    def is_equal(self, x1, x2) -> bool:  # +book
        """
        Check if the contents of two instances of the data structure are equal.

        :param x1: the first instance
        :param x2: the second instance
        :return: `True` if the contents are equal, `False` otherwise
        """

    def validate(self, x) -> None:  # +book
        """
        Check whether a given point in the space is valid.

        This function should be implemented such that it very carefully checks
        whether the argument `x` is a valid element of this space. It should
        check the Python data type of `x` and the type of its components and
        raise a `TypeError` if it does not match the exact requirements. It
        should also check the value of each element of `x` whether it is
        permitted. Once this function returns without throwing an exception,
        the user can rely on that the data structure `x` is correct.

        For example, if we have a space of
        :mod:`~moptipy.spaces.permutations` of the values from `1` to `n`,
        where the elements are represented as :class:`numpy.ndarray` objects,
        then this function should first check whether `x` is indeed an
        instance of :class:`numpy.ndarray`. If not, it should raise a
        `TypeError`. Then it could check whether the length of `x` is indeed
        `n` and raise a `ValueError` otherwise. It could then check whether
        each element of `x` is from `1..n` *and* occurs exactly one time (and
        otherwise raise a `ValueError`). Moreover, it should also check
        whether the numpy `dtype` of `x` is an appropriate integer data type
        and raise a `ValueError` otherwise. Hence, if this function checks an
        element `x` and does not raise any error, the user can rely on that
        this element `x` is, indeed, a fully valid permutation.

        In our system, we use this method for example at the end of
        optimization processes. Every solution that is written to a log file
        must pass through this method. In other words, we ensure that only
        valid solutions are stored. If the optimization algorithm or a search
        operator has a bug and sometimes may produce invalid data structures,
        this a) helps us in finding the bug and b) prevents us from storing
        invalid solutions. It is also strongly encouraged that the
        :meth:`from_str` method of any :class:`Space` implementation should
        run its results through :meth:`validate`. Since :meth:`from_str` can
        be used to, e.g., parse the data from the result sections of log
        files, this ensures that no corrupted or otherwise invalid data is
        parsed into an application.
        See https://thomasweise.github.io/moptipy/#data-formats for more
        information on log files.

        :param x: the point
        :raises TypeError: if the point `x` (or one of its elements, if
            applicable) has the wrong data type
        :raises ValueError: if the point `x` is invalid and/or simply is not
            an element of this space
        """

    def n_points(self) -> int:
        """
        Get the approximate number of different elements in the space.

        This operation can help us when doing tests of the space API
        implementations. If we know how many points exist in the space,
        then we can judge whether a method that randomly generates
        points is sufficiently random, for instance.

        By default, this method simply returns `2`. If you have a better
        approximation of the size of the space, then you should override it.

        :return: the approximate scale of the space
        """
        return 2


def check_space(space: Any, none_is_ok: bool = False) -> Space | None:
    """
    Check whether an object is a valid instance of :class:`Space`.

    :param space: the object
    :param none_is_ok: is it ok if `None` is passed in?
    :return: the object
    :raises TypeError: if `space` is not an instance of
        :class:`~moptipy.api.space.Space`

    >>> check_space(Space())
    Space
    >>> check_space(Space(), True)
    Space
    >>> check_space(Space(), False)
    Space
    >>> try:
    ...     check_space('A')
    ... except TypeError as te:
    ...     print(te)
    space should be an instance of moptipy.api.space.\
Space but is str, namely 'A'.
    >>> try:
    ...     check_space('A', True)
    ... except TypeError as te:
    ...     print(te)
    space should be an instance of moptipy.api.space.\
Space but is str, namely 'A'.
    >>> try:
    ...     check_space('A', False)
    ... except TypeError as te:
    ...     print(te)
    space should be an instance of moptipy.api.space.\
Space but is str, namely 'A'.
    >>>
    >>> try:
    ...     check_space(None)
    ... except TypeError as te:
    ...     print(te)
    space should be an instance of moptipy.api.space.\
Space but is None.
    >>> print(check_space(None, True))
    None
    >>> try:
    ...     check_space(None, False)
    ... except TypeError as te:
    ...     print(te)
    space should be an instance of moptipy.api.space.\
Space but is None.
    """
    if isinstance(space, Space):
        return space
    if none_is_ok and (space is None):
        return None
    raise type_error(space, "space", Space)



"""Make an LaTeX end-statistics table with column wrapping."""

from io import TextIOBase
from math import inf
from typing import Any, Callable, Final, Iterable, TextIO

from pycommons.io.console import logger
from pycommons.io.path import line_writer

from moptipy.evaluation.end_statistics import EndStatistics
from moptipy.evaluation.end_statistics import getter as es_getter
from moptipy.utils.number_renderer import default_get_int_renderer

#: the integer number renderer
__INT_2_STR: Final[Callable[[int], str]] = default_get_int_renderer()


def tabulate_end_stats(
        data: Iterable[EndStatistics],
        dest: Callable[[int], TextIO | TextIOBase],
        n_wrap: int = 3,
        max_rows: int = 50,
        stats: Iterable[tuple[Callable[[
            EndStatistics], int | float | None], str, bool,
        Callable[[int | float], str]]] = ((
        es_getter("bestF.mean"),
        r"\bestFmean", True,
        lambda v: __INT_2_STR(int(round(v)))), ),
        instance_get: Callable[[EndStatistics], str] =
        lambda es: es.instance,
        instance_sort_key: Callable[[str], Any] = lambda x: x,
        instance_name: Callable[[str], str] =
        lambda x: f"{{\\instStyle{{{x}}}}}",
        algorithm_get: Callable[[EndStatistics], str] = lambda x: x.algorithm,
        algorithm_sort_key: Callable[[str], Any] = lambda x: x,
        algorithm_name: Callable[[str], str] = lambda x: x,
        instance_cols: Iterable[tuple[str, Callable[[str], str]]] = (),
        best_format: Callable[[str], str] =
        lambda x: f"{{\\textbf{{{x}}}}}",
        instance_header: str = "instance",
        best_count_header: str | None = r"\nBest") -> None:
    """
    Make a table of end statistics that can wrap multiple pages, if need be.

    :param data: the source data
    :param dest: the destination generator
    :param n_wrap: the number of times we can wrap a table
    :param max_rows: the maximum rows per destination
    :param stats: the set of statistics: tuples of statistic, title, whether
        minimization or maximization, and a to-string converter
    :param instance_get: get the instance identifier
    :param instance_name: get the instance name, as it should be printed
    :param instance_sort_key: get the sort key for the instance
    :param algorithm_get: get the algorithm identifier
    :param algorithm_name: get the algorithm name, as it should be printed
    :param algorithm_sort_key: get the sort key for the algorithm
    :param instance_cols: the fixed instance columns
    :param best_format: format the best value
    :param instance_header: the header for the instance
    :param best_count_header: the header for the best count
    """
    if not isinstance(data, list):
        data = list(data)
    if list.__len__(data) <= 0:
        raise ValueError("Empty data?")
    if not isinstance(stats, list):
        stats = list(stats)
    n_stats: Final[int] = list.__len__(stats)
    if n_stats <= 0:
        raise ValueError("No statistics data?")

    # We create a map of data
    datamap: Final[dict[str, dict[str, EndStatistics]]] = {}
    for es in data:
        algorithm: str = algorithm_get(es)
        instance: str = instance_get(es)
        if instance in datamap:
            datamap[instance][algorithm] = es
        else:
            datamap[instance] = {algorithm: es}

    instances: Final[list[str]] = sorted(
        {es.instance for es in data}, key=instance_sort_key)
    if list.__len__(instances) <= 0:
        raise ValueError("No instance!")

    algorithms: Final[list[str]] = sorted({
        k for x in datamap.values() for k in x}, key=algorithm_sort_key)
    n_algorithms: Final[int] = list.__len__(algorithms)
    if n_algorithms <= 0:
        raise ValueError("No algorithms!")
    n_data_cols: Final[int] = n_algorithms * n_stats

    if not isinstance(instance_cols, tuple):
        instance_cols = tuple(instance_cols)
    n_inst_cols: Final[int] = tuple.__len__(instance_cols)

    # Now we compute all the values
    output: list[list[str]] = []

    best_count: Final[list[int]] = [0] * n_data_cols
    for instance in instances:
        row: list[str] = [instance_name(instance)]
        row.extend(ic[1](instance) for ic in instance_cols)
        values: list[int | float | None] = [None] * n_data_cols
        best: list[int | float] = [
            (inf if stat[2] else -inf) for stat in stats]

        # get the values
        idx: int = 0
        for algorithm in algorithms:
            es = datamap[instance][algorithm]
            for si, stat in enumerate(stats):
                values[idx] = value = stat[0](es)
                idx += 1
                if (value is not None) and ((stat[2] and (
                        value < best[si])) or ((not stat[2]) and (
                        value > best[si]))):
                    best[si] = value

        # format the values
        idx = 0
        for _ in range(n_algorithms):
            for si, stat in enumerate(stats):
                value = values[idx]
                idx += 1
                if value is None:
                    row.append("")
                    continue
                printer: str = stat[3](value)
                if ((stat[2] and (value <= best[si])) or ((not stat[2]) and (
                        value >= best[si]))):
                    best_count[idx - 1] += 1
                    printer = best_format(printer)
                row.append(printer)
        output.append(row)

    if best_count_header is not None:
        nc: int = n_inst_cols + 1
        row = [f"\\multicolumn{{{nc}}}{{r}}{{{best_count_header}}}"]
        best = [-1] * n_stats

        # get the best best counts
        idx = 0
        for _ in range(n_algorithms):
            for si in range(n_stats):
                value = best_count[idx]
                best[si] = max(value, best[si])
                idx += 1

        # format the best count values
        idx = 0
        for _ in range(n_algorithms):
            for si in range(n_stats):
                ivalue: int = best_count[idx]
                idx += 1
                printer = str(ivalue)
                if ivalue >= best[si]:
                    printer = best_format(printer)
                row.append(printer)
        output.append(row)

    # now we got all the data prepared and just need to arrange it
    n_output: Final[int] = list.__len__(output)
    pages: Final[list[list[list[int]]]] = []
    idx = 0
    keep_going: bool = True
    while keep_going:
        current_page: list[list[int]] = []
        pages.append(current_page)
        for _ in range(max_rows):
            current_row: list[int] = []
            current_page.append(current_row)
            for _ in range(n_wrap):
                current_row.append(idx)
                idx += 1
                if idx >= n_output:
                    keep_going = False
                    break
            if not keep_going:
                break
    # ok, now we know the data that we can put on each page
    # we now re-arrange it
    for current_page in pages:
        ids: list[int] = sorted(x for y in current_page for x in y)
        idx = 0
        for col in range(n_wrap):
            for current_row in current_page:
                if col < list.__len__(current_row):
                    current_row[col] = ids[idx]
                    idx += 1

    # now we construct the header and footer that should go into each
    # partial table
    header: list[str] = []
    footer: list[str] = [r"\hline%", r"\end{tabular}%"]

    writer: list[str] = ["l"]
    writer.extend("r" for _ in range(n_inst_cols))
    writer.extend(["r"] * n_stats * n_algorithms)
    txt: str = "".join(writer)
    header.append(f"\\begin{{tabular}}{{{'|'.join([txt] * n_wrap)}}}%")
    header.append(r"\hline%")

    writer.clear()
    writer.append(instance_header)  # noqa: PIE799
    writer.extend(f"\\multicolumn{{1}}{{c}}{{{ic[0]}}}"
                  for ic in instance_cols)
    writer.extend(f"\\multicolumn{{{n_stats}}}{{c}}{{{algorithm_name(algo)}}}"
                  for algo in algorithms)
    txt = "&".join(writer)
    if n_wrap > 1:
        writer[-1] = writer[-1].replace("{c}", "{c|}")
        txt_inner = "&".join(writer)
        header.append(f"{'&'.join([txt_inner] * (n_wrap - 1))}&{txt}\\\\%")
    else:
        header.append(f"{txt}\\\\%")

    writer.clear()
    if n_stats > 1:
        writer.clear()
        writer.append("")
        writer.extend("" for _ in range(n_inst_cols))
        writer.extend(f"\\multicolumn{{1}}{{c}}{{{sss[2]}}}"
                      for _ in range(n_algorithms) for sss in stats)
        txt = "&".join(writer)
        if n_wrap > 1:
            writer[-1] = writer[-1].replace("{c}", "{c|}")
            txt_inner = "&".join(writer)
            header.append(f"{'&'.join([txt_inner] * (n_wrap - 1))}&{txt}\\\\%")
        else:
            header.append(f"{txt}\\\\%")
        writer.clear()
    header.append(r"\hline%")

    for page_idx, current_page in enumerate(pages):
        logger(f"Now tacking data part {page_idx + 1}.")
        with dest(page_idx + 1) as stream:
            out_stream = line_writer(stream)
            for txt in header:
                out_stream(txt)
            for current_row in current_page:
                line: list[str] = [x for y in current_row for x in output[y]]
                out_stream(f"{'&'.join(line)}\\\\%")
            for txt in footer:
                out_stream(txt)



"""
The one-dimensional Ising problem.

The one-dimensional Ising problem describes a ring. For each bit that differs
from its (right-side) neighboring bit, a penalty of 1 is incurred. The optimum
is a bit string of either all ones or all zeros.

1. Simon Fischer and Ingo Wegener. The one-dimensional Ising model: Mutation
   versus recombination. *Theoretical Computer Science.* 344(2-3):208-225.
   November 2005. doi: https://doi.org/10.1016/j.tcs.2005.04.002
2. Carola Doerr and Furong Ye and Naama Horesh and Hao Wang and Ofer M. Shir
   and Thomas Bäck. Benchmarking Discrete Optimization Heuristics with
   IOHprofiler. Applied Soft Computing 88:106027, March 2020,
   doi: https://doi.org/10.1016/j.asoc.2019.106027.
3. Clarissa Van Hoyweghen, David Edward Goldberg, and Bart Naudts. From Twomax
   To The Ising Model: Easy And Hard Symmetrical Problems. *In Proceedings of
   the Genetic and Evolutionary Computation Conference (GECCO'02),* July 9-13,
   2002, New York, NY, USA, pages 626-633. Morgan Kaufmann.
   http://gpbib.cs.ucl.ac.uk/gecco2002/GA013.pdf
"""

import numba  # type: ignore
import numpy as np

from moptipy.examples.bitstrings.bitstring_problem import BitStringProblem


@numba.njit(nogil=True, cache=True)
def ising1d(x: np.ndarray) -> int:
    """
    Compute the objective value of the 1-dimensional Ising problem.

    :param x: the np array
    :return: the trap function value

    >>> print(ising1d(np.array([True, True, True, True, True])))
    0
    >>> print(ising1d(np.array([False, False, False, False, False])))
    0
    >>> print(ising1d(np.array([False, False, False, True, False])))
    2
    >>> print(ising1d(np.array([True, False, False, False, False])))
    2
    >>> print(ising1d(np.array([False, False, False, False, True])))
    2
    >>> print(ising1d(np.array([True, False, False, False, True])))
    2
    >>> print(ising1d(np.array([True, False, True, False, False])))
    4
    >>> print(ising1d(np.array([True, False, True, False, True, False])))
    6
    """
    s: int = 0
    prev: bool = x[-1]
    for cur in x:
        if cur != prev:
            s += 1
        prev = cur
    return s


class Ising1d(BitStringProblem):
    """The one-dimensional Ising problem."""

    def __init__(self, n: int) -> None:  # +book
        """
        Initialize the one-dimensional Ising problem.

        :param n: the dimension of the problem

        >>> print(Ising1d(7).n)
        7
        >>> print(Ising1d(3).evaluate(np.array([True, False, True])))
        2
        """
        super().__init__(n)
        self.evaluate = ising1d  # type: ignore

    def __str__(self) -> str:
        """
        Get the name of the one-dimensional Ising problem.

        :return: `ising1d_` + length of string

        >>> print(Ising1d(5))
        ising1d_5
        """
        return f"ising1d_{self.n}"



"""
Example for bit-string based objective functions.

The problems noted here are mainly well-known benchmark problems from the
discrete optimization community. For these problems, we designed the dedicated
base class :class:`~moptipy.examples.bitstrings.bitstring_problem.\
BitStringProblem`.

The following benchmark problems are provided:

1. :mod:`~moptipy.examples.bitstrings.ising1d`, the one-dimensional
   Ising model, where the goal is that all bits should have the same value as
   their neighbors.
2. The :mod:`~moptipy.examples.bitstrings.jump` problem is equivalent
   to :mod:`~moptipy.examples.bitstrings.onemax`, but has a deceptive
   region right before the optimum.
3. The :mod:`~moptipy.examples.bitstrings.leadingones` problem,
   where the goal is to find a bit string with the maximum number of leading
   ones.
4. The :mod:`~moptipy.examples.bitstrings.onemax` problem, where the
   goal is to find a bit string with the maximum number of ones.
5. The :mod:`~moptipy.examples.bitstrings.trap` problem, which is like
   OneMax, but with the optimum and worst-possible solution swapped. This
   problem is therefore highly deceptive.
6. The :mod:`~moptipy.examples.bitstrings.w_model`, a benchmark
   problem with tunable epistasis, uniform neutrality, and
   ruggedness/deceptiveness.
7. The :mod:`~moptipy.examples.bitstrings.zeromax` problem, where the
   goal is to find a bit string with the maximum number of zeros. This is the
   opposite of the OneMax problem.
"""



"""Simple routines for handling numbers and numerical stuff."""

import contextlib
from math import exp, inf, isfinite, isqrt, log, log2, nextafter
from typing import Final

from pycommons.types import type_error

#: The positive limit for doubles that can be represented exactly as ints.
DBL_INT_LIMIT_P: Final[float] = 9007199254740992.0  # = 1 << 53
#: The negative  limit for doubles that can be represented exactly as ints.
_DBL_INT_LIMIT_N: Final[float] = -DBL_INT_LIMIT_P


def __try_int(val: float) -> int | float:
    """
    Convert a float to an int without any fancy checks.

    :param val: the flot
    :returns: the float or int

    >>> from math import inf, nan, nextafter
    >>> type(__try_int(0.0))
    <class 'int'>
    >>> type(__try_int(0.5))
    <class 'float'>
    >>> type(__try_int(inf))
    <class 'float'>
    >>> type(__try_int(-inf))
    <class 'float'>
    >>> type(__try_int(nan))
    <class 'float'>
    >>> 1 << 53
    9007199254740992
    >>> type(__try_int(9007199254740992.0))
    <class 'int'>
    >>> __try_int(9007199254740992.0)
    9007199254740992
    >>> too_big = nextafter(9007199254740992.0, inf)
    >>> print(too_big)
    9007199254740994.0
    >>> type(__try_int(too_big))
    <class 'float'>
    >>> type(__try_int(-9007199254740992.0))
    <class 'int'>
    >>> __try_int(-9007199254740992.0)
    -9007199254740992
    >>> type(__try_int(-too_big))
    <class 'float'>
    """
    if _DBL_INT_LIMIT_N <= val <= DBL_INT_LIMIT_P:
        a = int(val)
        if a == val:
            return a
    return val


def try_int(val: int | float) -> int | float:
    """
    Attempt to convert a float to an integer.

    This method will convert a floating point number to an integer if the
    floating point number was representing an exact integer. This is the
    case if it has a) no fractional part and b) is in the range
    `-9007199254740992...9007199254740992`, i.e., the range where `+1` and
    `-1` work without loss of precision.

    :param val: the input value, which must either be `int` or `float`
    :return: an `int` if `val` can be represented as `int` without loss of
        precision, `val` otherwise
    :raises TypeError: if `val` is neither an instance of `int` nor of `float`
    :raises ValueError: if `val` is a `float`, but not finite

    >>> print(type(try_int(10.5)))
    <class 'float'>
    >>> print(type(try_int(10)))
    <class 'int'>

    >>> from math import inf, nan, nextafter
    >>> type(try_int(0.0))
    <class 'int'>
    >>> type(try_int(0.5))
    <class 'float'>
    >>> try:
    ...     try_int(inf)
    ... except ValueError as ve:
    ...     print(ve)
    Value must be finite, but is inf.
    >>> try:
    ...     try_int(-inf)
    ... except ValueError as ve:
    ...     print(ve)
    Value must be finite, but is -inf.
    >>> try:
    ...     try_int(nan)
    ... except ValueError as ve:
    ...     print(ve)
    Value must be finite, but is nan.
    >>> try:
    ...     try_int("blab")  # noqa  # type: off
    ... except TypeError as te:
    ...     print(te)
    val should be an instance of any in {float, int} but is str, namely \
'blab'.
    >>> type(try_int(9007199254740992.0))
    <class 'int'>
    >>> try_int(9007199254740992.0)
    9007199254740992
    >>> too_big = nextafter(9007199254740992.0, inf)
    >>> print(too_big)
    9007199254740994.0
    >>> type(try_int(too_big))
    <class 'float'>
    >>> type(try_int(-9007199254740992.0))
    <class 'int'>
    >>> try_int(-9007199254740992.0)
    -9007199254740992
    >>> type(try_int(-too_big))
    <class 'float'>
    """
    if isinstance(val, int):
        return val
    if isinstance(val, float):
        if not isfinite(val):
            raise ValueError(f"Value must be finite, but is {val}.")
        if _DBL_INT_LIMIT_N <= val <= DBL_INT_LIMIT_P:
            a = int(val)
            if a == val:
                return a
        return val
    raise type_error(val, "val", (int, float))


def try_int_div(a: int, b: int) -> int | float:
    """
    Try to divide two integers at best precision.

    Floating point divisions can incur some loss of precision. We try
    to avoid this here as much as possible. First, we check if `a` is
    divisible by `b` without any fractional part. If this is true, then
    we can do a pure integer division without loss of any precision.

    Otherwise, we would do a floating point division. This will lead the
    values be converted to floating point numbers, which can incur
    a loss of precision. In the past, we tried to divide both numbers by
    their `gcd` to make them smaller in order to avoid a loss of precision.
    But it seems that Python is already doing something like this internally
    when performing the `a / b` division anyway, so we don't do this anymore.
    However, it is still attempted to convert the result back to an integer.

    :param a: the first integer
    :param b: the second integer
    :return: a/b, either as `int` or as `float` but always a finite value
    :raises OverflowError: if the division must be performed using floating
        point arithmetic, but the result would be too large to fit into a
        `float`
    :raises ZeroDivisionError: if `b==0`

    >>> print(try_int_div(10, 2))
    5
    >>> print(try_int_div(10, -2))
    -5
    >>> print(try_int_div(-10, 2))
    -5
    >>> print(try_int_div(-10, -2))
    5
    >>> print(type(try_int_div(10, 2)))
    <class 'int'>
    >>> print(try_int_div(10, 3))
    3.3333333333333335
    >>> print(try_int_div(-10, 3))
    -3.3333333333333335
    >>> print(try_int_div(10, -3))
    -3.3333333333333335
    >>> print(try_int_div(-10, -3))
    3.3333333333333335
    >>> print(type(try_int_div(10, 3)))
    <class 'float'>
    >>> print(try_int_div(9007199254740992, 1))
    9007199254740992
    >>> print(try_int_div(2109792310235001520128, 234234))
    9007199254740992
    >>> print(try_int_div(2109792310235001520128, 234235))
    9007160801054503
    >>> print(try_int_div(2109792310235001520128, 234233))
    9007237708755818.0
    >>> large = 123456789012345678901234567890123456789012345678901234567\
89012345678901234567890123456789012345678901234567890123456789012345678901234\
56789012345678901234567890123456789012345678901234567890123456789012345678901\
23456789012345678901234567890123456789012345678901234567890123456789012345678\
90123456789012345678901234567890123456789012345678901234567890123456789012345\
67890123456789012345678901234567890123456789012345678901234567890123456789012\
3456789012345678901234567890123456789012345678901234567890123456789012345678\
90123456789012345678901234567890123456789012345678901234567890123456789012345\
678901234567890123456789012345678901234567890
    >>> try:
    ...     large / 1
    ... except OverflowError as oe:
    ...     print(oe)
    integer division result too large for a float
    >>> try_int_div(large, 1)
    123456789012345678901234567890123456789012345678901234567\
89012345678901234567890123456789012345678901234567890123456789012345678901234\
56789012345678901234567890123456789012345678901234567890123456789012345678901\
23456789012345678901234567890123456789012345678901234567890123456789012345678\
90123456789012345678901234567890123456789012345678901234567890123456789012345\
67890123456789012345678901234567890123456789012345678901234567890123456789012\
3456789012345678901234567890123456789012345678901234567890123456789012345678\
90123456789012345678901234567890123456789012345678901234567890123456789012345\
678901234567890123456789012345678901234567890
    >>> try_int_div(large * 7, 1 * 7)
    123456789012345678901234567890123456789012345678901234567\
89012345678901234567890123456789012345678901234567890123456789012345678901234\
56789012345678901234567890123456789012345678901234567890123456789012345678901\
23456789012345678901234567890123456789012345678901234567890123456789012345678\
90123456789012345678901234567890123456789012345678901234567890123456789012345\
67890123456789012345678901234567890123456789012345678901234567890123456789012\
3456789012345678901234567890123456789012345678901234567890123456789012345678\
90123456789012345678901234567890123456789012345678901234567890123456789012345\
678901234567890123456789012345678901234567890
    >>> try:
    ...     try_int_div(large, 7)
    ... except OverflowError as oe:
    ...     print(oe)
    integer division result too large for a float
    >>> try:
    ...     try_int_div(0, 0)
    ... except ZeroDivisionError as zde:
    ...     print(zde)
    integer division or modulo by zero
    >>> try:
    ...     try_int_div(1, 0)
    ... except ZeroDivisionError as zde:
    ...     print(zde)
    integer division or modulo by zero
    >>> try:
    ...     try_int_div(-1, 0)
    ... except ZeroDivisionError as zde:
    ...     print(zde)
    integer division or modulo by zero
    """
    # First try pure integer division, in case `a` is divisible by `b`.
    int_div_test: Final[int] = a // b
    if (int_div_test * b) == a:
        return int_div_test
    # If that does not work, use the floating point division.
    return __try_int(a / b)


def try_float_div(a: int | float, b: int | float) -> int | float:
    """
    Try to divide two numbers at best precision.

    First, we will check if we can convert the numbers to integers
    without loss of precision via :func:`try_int`. If yes, then
    we go for the maximum-precision integer division via :func:`try_int_div`.
    If no, then we do the normal floating point division and try to convert
    the result to an integer if that can be done without loss of precision.

    :param a: the first number
    :param b: the second number
    :return: `a/b`, but always finite

    :raises ValueError: if either one of the arguments or the final result
        would not be finite

    >>> try_float_div(1e180, 1e60)
    1.0000000000000001e+120
    >>> try_float_div(1e60, 1e-60)
    1e+120
    >>> try_float_div(1e14, 1e-1)
    1000000000000000
    >>> try_float_div(1e14, -1e-1)
    -1000000000000000
    >>> try_float_div(-1e14, 1e-1)
    -1000000000000000
    >>> try_float_div(-1e14, -1e-1)
    1000000000000000
    >>> try_float_div(1e15, 1e-1)
    1e+16
    >>> try_float_div(1e15, -1e-1)
    -1e+16
    >>> try_float_div(-1e15, 1e-1)
    -1e+16
    >>> try_float_div(-1e15, -1e-1)
    1e+16
    >>> try_float_div(1e15, 1e-15)
    9.999999999999999e+29

    >>> print(type(try_float_div(10, 2)))
    <class 'int'>
    >>> print(type(try_float_div(10, 3)))
    <class 'float'>
    >>> print(type(try_float_div(10, 0.5)))
    <class 'int'>

    >>> from math import inf, nan
    >>> try:
    ...     try_float_div(1.0, 0)
    ... except ZeroDivisionError as zde:
    ...     print(zde)
    integer division or modulo by zero

    >>> try:
    ...     try_float_div(1.0, -0.0)
    ... except ZeroDivisionError as zde:
    ...     print(zde)
    integer division or modulo by zero

    >>> try:
    ...     try_float_div(inf, 0)
    ... except ValueError as ve:
    ...     print(ve)
    Value must be finite, but is inf.

    >>> try:
    ...     try_float_div(-inf, 0)
    ... except ValueError as ve:
    ...     print(ve)
    Value must be finite, but is -inf.

    >>> try:
    ...     try_float_div(nan, 0)
    ... except ValueError as ve:
    ...     print(ve)
    Value must be finite, but is nan.

    >>> try:
    ...     try_float_div(1, inf)
    ... except ValueError as ve:
    ...     print(ve)
    Value must be finite, but is inf.

    >>> try:
    ...     try_float_div(1, -inf)
    ... except ValueError as ve:
    ...     print(ve)
    Value must be finite, but is -inf.

    >>> try:
    ...     try_float_div(1, nan)
    ... except ValueError as ve:
    ...     print(ve)
    Value must be finite, but is nan.

    >>> try:
    ...     try_float_div(1e300, 1e-60)
    ... except ValueError as ve:
    ...     print(ve)
    Result must be finite, but is 1e+300/1e-60=inf.
    """
    ia: Final[int | float] = try_int(a)
    ib: Final[int | float] = try_int(b)
    if isinstance(ia, int) and isinstance(ib, int):
        return try_int_div(ia, ib)
    val: Final[float] = ia / ib
    if not isfinite(val):
        raise ValueError(f"Result must be finite, but is {a}/{b}={val}.")
    return __try_int(val)


def __bin_search_root(value: int, power: int) -> int:
    """
    Search a truncated integer root via binary search.

    :param value: the integer value to compute the root of
    :param power: the integer power of the root
    :returns: the integer root, truncated if necessary
    """
    int_sqrt: Final[int] = isqrt(value)
    if power == 2:
        return int_sqrt

    # compute the maximum base root
    root_min: int = 1
    root_max: int = (int_sqrt - 1) if value > 3 else 1
    while root_max >= root_min:
        root_mid = isqrt(root_min * root_max)
        power_mid = root_mid ** power
        if power_mid > value:
            root_max = root_mid - 1
        elif power_mid < value:
            root_min = root_mid + 1
        else:
            return root_mid
    return root_max


def try_int_root(value: int, power: int,
                 none_on_overflow: bool = True) -> int | float | None:
    """
    Compute `value**(1/power)` where `value` and `power` are both integers.

    :param value: the integer value to compute the root of
    :param power: the integer power of the root
    :param none_on_overflow: `True` if `None` should be returned if we
        encounter an :class:`OverflowError`, `False` if we should simply
        re-raise it
    :returns: the root, or `None` if we encounter an overflow

    >>> try_int_root(100, 2)
    10
    >>> try_int_root(100, 3) - (100 ** (1/3))
    0.0
    >>> try_int_root(123100, 23) - (123100 ** (1/23))
    0.0
    >>> 123100**1000 - try_int_root(123100**1000, 1000)**1000
    0
    >>> 11**290 - try_int_root(11**290, 290)**290
    0
    >>> (abs(1.797E308 - try_int_root(int(1.797E308), 10) ** 10)
    ...     < abs(1.797E308 - (int(1.797E308) ** 0.1) ** 10))
    True
    """
    if not isinstance(value, int):
        raise type_error(value, "value", int)
    if not isinstance(power, int):
        raise type_error(power, "power", int)
    if power <= 0:
        raise ValueError(f"power must be positive but is {power}.")
    if value <= 1:
        if value < 0:
            raise ValueError(f"value must not be negative, but is {value}.")
        return value
    if power == 1:
        return value

    # first, approximate root with crude binary search
    int_root: int = __bin_search_root(value, power)
    root_power: int = int_root ** power
    if root_power >= value:
        if root_power == value:
            return int_root  # ok, we are done
        raise ValueError(f"{int_root}**{power}={root_power} > {value}?")

    # Now try to remove integer factors from the value and the root,
    # to achieve a more accurate division of the rest.
    # This is probably extremely inefficient.
    root_base: int = 1
    i: int = 2
    end: int = min(10_000, int_root)
    while i < end:
        div: int = i ** power
        if (value % div) == 0:
            root_base *= i
            value //= div
            int_root = (int_root + i - 1) // i
            end = min(end, int_root)
        else:
            i += 1

    # value is now reduced by any integer factor that we can pull out
    # of the root. These factors are aggregated in root_base.
    # If root_base != 1, we need to re-compute the root of the remainder
    # inside value. Whatever root we now compute of value, we must multiply
    # it with root_base.
    if root_base != 1:
        int_root = __bin_search_root(value, power)

    # check again
    root_power = int_root ** power
    if root_power >= value:
        if root_power == value:
            return root_base * int_root  # ok, we are done
        raise ValueError(f"{int_root}**{power}={root_power} > {value}?")

    # from now on, root may be either and int or (more likely) a float.
    root: int | float = int_root
    try:
        rest: int | float = try_int_div(value, root_power)
        rest_root: int | float = __try_int(rest ** (1.0 / power))
        root = __try_int(root * rest_root)
    except OverflowError:  # as ofe:
        if none_on_overflow:
            return None
        raise  # raises ofe again

    # OK, we got an approximate root of what remains of value.
    # Let's see if we can refine it.
    with contextlib.suppress(OverflowError):
        diff = abs((root ** power) - value)
        root2: int | float = __try_int(exp(log(value) / root))
        diff2: int | float = abs((root2 ** power) - value)
        if diff2 < diff:
            diff = diff2
            root = root2
        root2 = __try_int(2 ** (log2(value) / root))
        diff2 = abs((root2 ** power) - value)
        if diff2 < diff:
            diff = diff2
            root = root2

        rdn = root
        rup = root
        while True:
            rdn = nextafter(rdn, -inf)
            apd = abs((rdn ** power) - value)
            if apd > diff:
                break
            diff = apd
            root = rdn
        while True:
            rup = nextafter(rup, inf)
            apd = abs((rup ** power) - value)
            if apd > diff:
                break
            diff = apd
            root = rup

    return root_base * __try_int(root)



"""A set of process wrappers that can be used for surrogate modeling."""

from math import inf, isfinite
from typing import Any, Callable, Final

import numpy as np

from moptipy.api.process import Process


class _Surrogate(Process):
    """A surrogate process."""

    def __init__(self, owner: Process, max_fes: int):
        super().__init__()
        #: the owning process
        self._owner: Final[Process] = owner
        self.get_random = owner.get_random  # type: ignore
        a = owner.get_consumed_time_millis  # type: ignore
        self.get_consumed_time_millis = a  # type: ignore
        a = owner.get_max_time_millis  # type: ignore
        self.get_max_time_millis = a  # type: ignore
        a = owner.get_last_improvement_time_millis  # type: ignore
        self.get_last_improvement_time_millis = a  # type: ignore
        self.has_log = owner.has_log  # type: ignore
        self.add_log_section = owner.add_log_section  # type: ignore
        self.lower_bound = owner.lower_bound  # type: ignore
        self.upper_bound = owner.upper_bound  # type: ignore
        self.create = owner.create  # type: ignore
        self.copy = owner.copy  # type: ignore
        self.to_str = owner.to_str  # type: ignore
        self.is_equal = owner.is_equal  # type: ignore
        self.from_str = owner.from_str  # type: ignore
        self.validate = owner.validate  # type: ignore
        self.n_points = owner.n_points  # type: ignore
        #: the maximum FEs
        self.max_fes: Final[int] = max_fes
        #: the FEs that we still have left
        self._fes_left: int = max_fes
        #: did we terminate?
        self._terminated: bool = False
        #: the fast call to the owner's should_terminate method
        self.__should_terminate: Final[Callable[[], bool]] \
            = owner.should_terminate

    def should_terminate(self) -> bool:
        return self._terminated or self.__should_terminate()

    def terminate(self) -> None:
        self._terminated = True

    def register(self, x, f: int | float) -> None:
        raise ValueError("you should not call this function.")

    def get_consumed_fes(self) -> int:
        return self.max_fes - self._fes_left

    def get_last_improvement_fe(self) -> int:
        return 1 if self._fes_left < self.max_fes else 0

    def get_max_fes(self) -> int:
        return self.max_fes

    def __str__(self) -> str:
        return f"{self.max_fes}_{self._owner}"


class _SurrogateApply(_Surrogate):
    """A process running for a `n` FEs and collecting the results."""

    #: the internal evaluation function
    _evaluate: Callable[[np.ndarray], np.ndarray]

    def __init__(self, owner: Process, max_fes: int) -> None:
        super().__init__(owner, max_fes)
        #: the best-so-far solution
        self._best_x: Final[np.ndarray] = owner.create()
        #: the best-so-far objective value
        self._best_f: int | float = inf

    def has_best(self) -> bool:
        return isfinite(self._best_f)

    def get_best_f(self) -> int | float:
        return self._best_f

    def get_copy_of_best_x(self, x) -> None:
        np.copyto(x, self._best_x)

    def evaluate(self, x) -> float | int:
        f: Final[float] = self._evaluate(
            x.reshape((1, x.shape[0])))[0]
        fel: Final[int] = self._fes_left - 1
        self._fes_left = fel
        if fel <= 0:
            self._terminated = True
        if isfinite(f):
            if f < self._best_f:
                self._best_f = f
                np.copyto(self._best_x, x)
            return f
        return inf


class _SurrogateWarmup(_Surrogate):
    """A process running for a `n` FEs and collecting the results."""

    def __init__(self, owner: Process, max_fes: int,
                 point_collector: Callable[[np.ndarray], None],
                 f_collector: Callable[[int | float], None]):
        super().__init__(owner, max_fes)
        #: the point collector
        self.__point_collector = point_collector
        #: the objective value collector
        self.__f_collector = f_collector
        self.has_best = owner.has_best  # type: ignore
        self.get_copy_of_best_x = owner.get_copy_of_best_x  # type: ignore
        self.get_best_f = owner.get_best_f  # type: ignore
        #: the owner's evaluation function
        self.__evaluate: Final[Callable[[Any], int | float]] = owner.evaluate

    def evaluate(self, x) -> float | int:
        f: Final[int | float] = self.__evaluate(x)
        self.__point_collector(x.copy())
        self.__f_collector(f)
        fel: Final[int] = self._fes_left - 1
        self._fes_left = fel
        if fel <= 0:
            self._terminated = True
        return f

    def lower_bound(self) -> float:
        return -inf

    def upper_bound(self) -> float:
        return inf

    def __str__(self) -> str:
        return f"surrogateWarmup_{super().__str__()}"



"""
The random sampling algorithm `rs`.

The random sampling algorithm keeps creating random solutions until
the computational budget is exhausted or any other termination
criterion is met.
In a loop that is repeated until the termination criterion
:meth:`~moptipy.api.process.Process.should_terminate` becomes `True`,
it

- applies the nullary search operator, an implementation of
  :meth:`~moptipy.api.operators.Op0.op0`, to sample exactly one single
  random solution. It then
- evaluates the solution by passing it to
  :meth:`~moptipy.api.process.Process.evaluate`.

This algorithm has a very bad performance. It makes no use of the
information that is learned during the search. Every single solution is
created completely independent on anything that has happened before
in the algorithm. Only if the problem is completely random, has no
structure, or is entirely deceptive, this algorithm can be of any use.

Basically, this algorithm is an iterated version of the `1rs` algorithm
implemented in
:class:`~moptipy.algorithms.single_random_sample.SingleRandomSample`.

1. Thomas Weise. *Optimization Algorithms.* 2021. Hefei, Anhui, China:
   Institute of Applied Optimization (IAO), School of Artificial Intelligence
   and Big Data, Hefei University. http://thomasweise.github.io/oa/
"""
from typing import Callable, Final

from numpy.random import Generator

from moptipy.api.algorithm import Algorithm0
from moptipy.api.operators import Op0
from moptipy.api.process import Process


# start book
class RandomSampling(Algorithm0):
    """In each step, random sampling creates a new, random solution."""

    def solve(self, process: Process) -> None:
        """
        Apply the random sampling approach to an optimization problem.

        :param process: the black-box process object
        """
        x: Final = process.create()  # Create the solution record.
        # Obtain the random number generator.
        random: Final[Generator] = process.get_random()

        # Put function references in variables to save time.
        evaluate: Final[Callable] = process.evaluate  # the objective
        op0: Final[Callable] = self.op0.op0  # the nullary operator
        should_terminate: Final[Callable] = process.should_terminate

        while not should_terminate():  # Until we need to quit...
            op0(random, x)  # Sample a completely random solution.
            evaluate(x)  # Evaluate the solution ... but ignore result.
# end book

    def __init__(self, op0: Op0) -> None:
        """
        Create the random sampling algorithm.

        :param op0: the nullary search operator
        """
        super().__init__("rs", op0)



"""A tool for writing a section with system information into log files."""
import contextlib
import importlib.metadata as ilm
import os
import platform
import re
import socket
import sys
from contextlib import suppress
from datetime import UTC, datetime
from typing import Final, Iterable

import psutil  # type: ignore
from pycommons.io.csv import CSV_SEPARATOR, SCOPE_SEPARATOR
from pycommons.io.path import Path
from pycommons.processes.caller import is_build

import moptipy.version as ver
from moptipy.api import logging
from moptipy.utils.logger import (
    KEY_VALUE_SEPARATOR,
    InMemoryLogger,
    KeyValueLogSection,
    Logger,
)


def __cpu_affinity(proc: psutil.Process | None = None) -> str | None:
    """
    Get the CPU affinity.

    :param proc: the process handle
    :return: the CPU affinity string.
    """
    if proc is None:
        proc = psutil.Process()
    if proc is None:
        return None
    cpua = proc.cpu_affinity()
    if cpua is not None:
        cpua = CSV_SEPARATOR.join(map(str, cpua))
        if len(cpua) > 0:
            return cpua
    return None


#: the dependencies
__DEPENDENCIES: set[str] | None = {
    "cmaes", "contourpy", "cycler", "fonttools", "intel-cmplr-lib-rt",
    "joblib", "kiwisolver", "llvmlite", "matplotlib", "moptipy", "numba",
    "numpy", "packaging", "pdfo", "Pillow", "psutil", "pycommons",
    "pyparsing", "python-dateutil", "scikit-learn", "scipy", "setuptools",
    "six", "threadpoolctl"}


def add_dependency(dependency: str,
                   ignore_if_make_build: bool = False) -> None:
    """
    Add a dependency so that its version can be stored in log files.

    Warning: You must add all dependencies *before* the first log file is
    written. As soon as the :func:`log_sys_info` is invoked for the first
    time, adding new dependencies will cause an error. And writing a log
    file via the :mod:`~moptipy.api.experiment` API or to a file specified in
    the :mod:`~moptipy.api.execution` API will invoke this function.

    :param dependency: the basic name of the library, exactly as you would
        `import` it in a Python module. For example, to include the version of
        `numpy` in the log files, you would do `add_dependency("numpy")` (of
        course, the version of `numpy` is already automatically included
        anyway).
    :param ignore_if_make_build: should this dependency be ignored if this
        method is invoked during a `make` build? This makes sense if the
        dependency itself is a package which is uninstalled and then
        re-installed during a `make` build process. In such a situation, the
        dependency version may be unavailable and cause an exception.
    :raises TypeError: if `dependency` is not a string
    :raises ValueError: if `dependency` is an invalid string or the log
        information has already been accessed before and modifying it now is
        not permissible.
    """
    if (str.__len__(dependency) <= 0) or (dependency != str.strip(dependency))\
            or (" " in dependency):
        raise ValueError(f"Invalid dependency string {dependency!r}.")
    if __DEPENDENCIES is None:
        raise ValueError(
            f"Too late. Cannot add dependency {dependency!r} anymore.")
    if ignore_if_make_build and is_build():
        return
    __DEPENDENCIES.add(dependency)


# noinspection PyBroadException
def __make_sys_info() -> str:
    """
    Build the system info string.

    This method is only used once and then deleted.

    :returns: the system info string.
    """
    global __DEPENDENCIES  # noqa: PLW0603  # pylint: disable=W0603
    if __DEPENDENCIES is None:
        raise ValueError("Cannot re-create log info.")
    dep: Final[set[str]] = __DEPENDENCIES
    __DEPENDENCIES = None  # noqa: PLW0603  # pylint: disable=W0603

    def __v(sec: KeyValueLogSection, key: str, value) -> None:
        """
        Create a key-value pair if value is not empty.

        :param sec: the section to write to.
        :param key: the key
        :param value: an arbitrary value, maybe consisting of multiple lines.
        """
        if value is None:
            return
        value = str.strip(" ".join([str.strip(ts) for ts in
                                    str.strip(str(value)).split("\n")]))
        if str.__len__(value) <= 0:
            return
        sec.key_value(key, value)

    def __get_processor_name() -> str | None:
        """
        Get the processor name.

        :returns: a string if there is any processor information
        """
        with contextlib.suppress(Exception):
            if platform.system() == "Windows":
                return platform.processor()
            if platform.system() == "Linux":
                with Path("/proc/cpuinfo").open_for_read() as rd:
                    for line in rd:
                        if "model name" in line:
                            return re.sub(
                                pattern=".*model name.*:",
                                repl="", string=line, count=1).strip()
        return None

    def __get_mem_size_sysconf() -> int | None:
        """
        Get the memory size information from sysconf.

        :returns: an integer with the memory size if available
        """
        with contextlib.suppress(Exception):
            k1 = "SC_PAGESIZE"
            if k1 not in os.sysconf_names:
                k1 = "SC_PAGE_SIZE"
                if k1 not in os.sysconf_names:
                    return None
            v1 = os.sysconf(k1)
            if not isinstance(v1, int):
                return None
            k2 = "SC_PHYS_PAGES"
            if k2 not in os.sysconf_names:
                k2 = "_SC_PHYS_PAGES"
                if k2 not in os.sysconf_names:
                    return None
            v2 = os.sysconf(k2)
            if not isinstance(v1, int):
                return None

            if (v1 > 0) and (v2 > 0):
                return v1 * v2
        return None

    def __get_mem_size_meminfo() -> int | None:
        """
        Get the memory size information from meminfo.

        :returns: an integer with the memory size if available
        """
        with contextlib.suppress(Exception):
            with Path("/proc/meminfo").open_for_read() as rd:
                meminfo = {i.split()[0].rstrip(":"): int(i.split()[1])
                           for i in rd}
            mem_kib = meminfo["MemTotal"]  # e.g. 3921852
            mem_kib = int(mem_kib)
            if mem_kib > 0:
                return 1024 * mem_kib
        return None

    def __get_mem_size() -> int | None:
        """
        Get the memory size information from any available source.

        :returns: an integer with the memory size if available
        """
        vs = __get_mem_size_sysconf()
        if vs is None:
            vs = __get_mem_size_meminfo()
        if vs is None:
            return psutil.virtual_memory().total
        return vs

    # log all information in memory to convert it to one constant string.
    with InMemoryLogger() as imr:
        with imr.key_values(logging.SECTION_SYS_INFO) as kv:
            with kv.scope(logging.SCOPE_SESSION) as k:
                __v(k, logging.KEY_SESSION_START,
                    datetime.now(tz=UTC))
                __v(k, logging.KEY_NODE_NAME, platform.node())
                proc = psutil.Process()
                __v(k, logging.KEY_PROCESS_ID, hex(proc.pid))
                cpua = __cpu_affinity(proc)
                if cpua is not None:
                    __v(k, logging.KEY_CPU_AFFINITY, cpua)
                del proc, cpua

                # get the command line and working directory of the process
                with contextlib.suppress(Exception):
                    proc = psutil.Process(os.getpid())
                    cmd = proc.cmdline()
                    if isinstance(cmd, Iterable):
                        __v(k, logging.KEY_COMMAND_LINE,
                            " ".join(map(repr, proc.cmdline())))
                    cwd = proc.cwd()
                    if isinstance(cwd, str):
                        __v(k, logging.KEY_WORKING_DIRECTORY,
                            repr(proc.cwd()))

                # see https://stackoverflow.com/questions/166506/.
                s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                try:
                    # doesn't even have to be reachable
                    s.connect(("10.255.255.255", 1))
                    ip = s.getsockname()[0]
                except Exception:  # noqa: BLE001,PIE786
                    ip = "127.0.0.1"
                finally:
                    s.close()
                __v(k, logging.KEY_NODE_IP, ip)

            with kv.scope(logging.SCOPE_VERSIONS) as k:
                for package in sorted(dep):
                    if package == "moptipy":
                        __v(k, "moptipy", ver.__version__)
                    else:
                        use_v: str | None = None
                        with suppress(BaseException):
                            use_v = str.strip(ilm.version(package))
                        if (use_v is None) or (str.__len__(use_v) <= 0):
                            use_v = "notFound"
                        __v(k, package.replace("-", ""), use_v)

            with kv.scope(logging.SCOPE_HARDWARE) as k:
                __v(k, logging.KEY_HW_MACHINE, platform.machine())
                __v(k, logging.KEY_HW_N_PHYSICAL_CPUS,
                    psutil.cpu_count(logical=False))
                __v(k, logging.KEY_HW_N_LOGICAL_CPUS,
                    psutil.cpu_count(logical=True))

                # store the CPU speed information
                cpuf: dict[tuple[int, int], int] = {}
                total: int = 0
                for cf in psutil.cpu_freq(True):
                    t = (int(cf.min), int(cf.max))
                    cpuf[t] = cpuf.get(t, 0) + 1
                    total += 1
                memlst: list[tuple[int, ...]]
                if total > 1:
                    memlst = [(key[0], key[1], value) for
                              key, value in cpuf.items()]
                    memlst.sort()
                else:
                    memlst = list(cpuf)

                def __make_mhz_str(tpl: tuple[int, ...]) -> str:
                    """Convert a MHz tuple to a string."""
                    base: str = f"({tpl[0]}MHz..{tpl[1]}MHz)" \
                        if tpl[1] > tpl[0] else f"{tpl[0]}MHz"
                    return base if (len(tpl) < 3) or (tpl[2] <= 1) \
                        else f"{base}*{tpl[2]}"
                __v(k, logging.KEY_HW_CPU_MHZ,
                    "+".join([__make_mhz_str(t) for t in memlst]))

                __v(k, logging.KEY_HW_BYTE_ORDER, sys.byteorder)
                __v(k, logging.KEY_HW_CPU_NAME, __get_processor_name())
                __v(k, logging.KEY_HW_MEM_SIZE, __get_mem_size())

            with kv.scope(logging.SCOPE_PYTHON) as k:
                __v(k, logging.KEY_PYTHON_VERSION, sys.version)
                __v(k, logging.KEY_PYTHON_IMPLEMENTATION,
                    platform.python_implementation())

            with kv.scope(logging.SCOPE_OS) as k:
                __v(k, logging.KEY_OS_NAME, platform.system())
                __v(k, logging.KEY_OS_RELEASE, platform.release())
                __v(k, logging.KEY_OS_VERSION, platform.version())

        lst = imr.get_log()

    if len(lst) < 3:
        raise ValueError("sys info turned out to be empty?")
    __DEPENDENCIES = None
    return "\n".join(lst[1:(len(lst) - 1)])


def get_sys_info() -> str:
    r"""
    Get the system information as string.

    :returns: the system information as string

    >>> raw_infos = get_sys_info()
    >>> raw_infos is get_sys_info()  # caching!
    True
    >>> for k in raw_infos.split("\n"):
    ...     print(k[:k.find(": ")])
    session.start
    session.node
    session.processId
    session.cpuAffinity
    session.commandLine
    session.workingDirectory
    session.ipAddress
    version.Pillow
    version.cmaes
    version.contourpy
    version.cycler
    version.fonttools
    version.intelcmplrlibrt
    version.joblib
    version.kiwisolver
    version.llvmlite
    version.matplotlib
    version.moptipy
    version.numba
    version.numpy
    version.packaging
    version.pdfo
    version.psutil
    version.pycommons
    version.pyparsing
    version.pythondateutil
    version.scikitlearn
    version.scipy
    version.setuptools
    version.six
    version.threadpoolctl
    hardware.machine
    hardware.nPhysicalCpus
    hardware.nLogicalCpus
    hardware.cpuMhz
    hardware.byteOrder
    hardware.cpu
    hardware.memSize
    python.version
    python.implementation
    os.name
    os.release
    os.version
    """
    the_object: Final[object] = get_sys_info
    the_attr: Final[str] = "__the_sysinfo"
    if hasattr(the_object, the_attr):
        return getattr(the_object, the_attr)
    sys_info: Final[str] = __make_sys_info()
    setattr(the_object, the_attr, sys_info)
    return sys_info


def update_sys_info_cpu_affinity() -> None:
    """Update the CPU affinity of the system information."""
    sys_info_str = get_sys_info()
    start = (f"\n{logging.SCOPE_SESSION}{SCOPE_SEPARATOR}"
             f"{logging.KEY_CPU_AFFINITY}{KEY_VALUE_SEPARATOR}")
    start_i = sys_info_str.find(start)
    if start_i < 0:
        return   # no affinity, don't need to update
    start_i += len(start)
    end_i = sys_info_str.find("\n", start_i)
    if end_i <= start_i:
        raise ValueError(f"Empty {logging.KEY_CPU_AFFINITY}?")
    affinity = __cpu_affinity()
    if affinity is None:
        raise ValueError(
            f"first affinity query is {sys_info_str[start_i:end_i]},"
            f" but second one is None?")
    sys_info_str = f"{sys_info_str[:start_i]}{affinity}{sys_info_str[end_i:]}"

    the_object: Final[object] = get_sys_info
    the_attr: Final[str] = "__the_sysinfo"
    setattr(the_object, the_attr, sys_info_str)


def log_sys_info(logger: Logger) -> None:
    r"""
    Write the system information section to a logger.

    The concept of this method is that we only construct the whole system
    configuration exactly once in memory and then always directly flush it
    as a string to the logger. This is much more efficient than querying it
    every single time.

    :param logger: the logger

    >>> from moptipy.utils.logger import InMemoryLogger
    >>> with InMemoryLogger() as l:
    ...     log_sys_info(l)
    ...     log = l.get_log()
    >>> print(log[0])
    BEGIN_SYS_INFO
    >>> print(log[-1])
    END_SYS_INFO
    >>> log[1:-1] == get_sys_info().split('\n')
    True
    """
    with logger.text(logging.SECTION_SYS_INFO) as txt:
        txt.write(get_sys_info())



"""The well-known sphere function."""

import numba  # type: ignore
import numpy as np

from moptipy.api.objective import Objective


@numba.njit(nogil=True, cache=True)
def sphere(x: np.ndarray) -> float:
    """
    Get the sum of the squares of the elements in an array.

    :param x: the np array
    :return: the sum of the squares of all elements

    >>> print(sphere(np.array([1.0, -2.0, 3.0])))
    14.0
    >>> print(sphere(np.array([0.0, 0.0, 0.0])))
    0.0
    """
    return float(np.sum(x ** 2))


class Sphere(Objective):
    """The well-known sphere function."""

    def __init__(self) -> None:
        """Initialize the sphere function."""
        self.evaluate = sphere  # type: ignore

    def lower_bound(self) -> float:
        """
        Get the lower bound of the sphere problem.

        :return: 0

        >>> print(Sphere().lower_bound())
        0.0
        """
        return 0.0

    def __str__(self) -> str:
        """
        Get the name of the sphere problem.

        :return: `sphere`
        :retval "sphere": always

        >>> print(Sphere())
        sphere
        """
        return "sphere"



"""Perform tests on the Job Shop Scheduling Problem."""

from typing import Any, Callable, Final, Iterable, cast

from numpy.random import Generator, default_rng
from pycommons.types import type_error

from moptipy.api.algorithm import Algorithm
from moptipy.api.mo_algorithm import MOAlgorithm
from moptipy.api.mo_problem import MOProblem
from moptipy.api.objective import Objective
from moptipy.examples.jssp.gantt import Gantt
from moptipy.examples.jssp.gantt_space import GanttSpace
from moptipy.examples.jssp.instance import Instance
from moptipy.examples.jssp.makespan import Makespan
from moptipy.examples.jssp.ob_encoding import OperationBasedEncoding
from moptipy.examples.jssp.worktime import Worktime
from moptipy.mo.problem.weighted_sum import WeightedSum
from moptipy.operators.permutations.op0_shuffle import Op0Shuffle
from moptipy.spaces.permutations import Permutations
from moptipy.tests.algorithm import validate_algorithm
from moptipy.tests.mo_algorithm import validate_mo_algorithm
from moptipy.tests.objective import validate_objective


def jssp_instances_for_tests() -> Iterable[str]:
    """
    Get a sequence of JSSP instances to test on.

    :returns: an iterable of JSSP instance names
    """
    r = default_rng()
    ri = r.integers
    insts: list[str] = [
        "demo", "ft06", "ft10", f"abz{ri(5, 10)}", f"dmu{ri(10, 81)}",
        f"orb0{ri(1, 10)}", f"swv{ri(10, 21)}", f"ta{ri(10, 65)}",
        f"ta{ri(65, 70)}", f"ta{ri(70, 75)}", f"yn{ri(1, 5)}"]
    r.shuffle(cast(list, insts))
    return insts


def make_gantt_valid(inst: Instance) -> Callable[[Generator, Gantt], Gantt]:
    """
    Make a function that creates valid Gantt charts.

    :param inst: the JSSP instance
    :returns: a function that can make gantt charts valid
    """
    pr = Permutations.with_repetitions(inst.jobs, inst.machines)
    op0 = Op0Shuffle(pr)
    oe = OperationBasedEncoding(inst)

    def __make_valid(prnd: Generator, x: Gantt, ppr=pr,
                     pop0=op0, poe=oe) -> Gantt:
        xx = ppr.create()
        pop0.op0(prnd, xx)
        poe.decode(xx, x)
        return x

    return __make_valid


def validate_algorithm_on_1_jssp(
        algorithm: Algorithm | Callable[
            [Instance, Permutations, Objective], Algorithm],
        instance: str | None = None, max_fes: int = 100,
        required_result: int | None = None,
        post: Callable[[Algorithm, int], Any] | None = None) -> None:
    """
    Check the validity of a black-box algorithm on the JSSP.

    :param algorithm: the algorithm or algorithm factory
    :param instance: the instance name, or `None` to randomly pick one
    :param max_fes: the maximum number of FEs
    :param required_result: the optional required result quality
    :param post: a check to run after each execution of the algorithm,
        receiving the algorithm and the number of consumed FEs as parameter
    """
    if not (isinstance(algorithm, Algorithm) or callable(algorithm)):
        raise type_error(algorithm, "algorithm", Algorithm, True)
    if instance is None:
        instance = str(default_rng().choice(Instance.list_resources()))
    if not isinstance(instance, str):
        raise type_error(instance, "JSSP instance name", (str, None))
    inst = Instance.from_resource(instance)
    if not isinstance(inst, Instance):
        raise type_error(inst, f"loaded JSSP instance {instance!r}", Instance)
    if (post is not None) and (not callable(post)):
        raise type_error(post, "post", None, call=True)

    search_space = Permutations.with_repetitions(inst.jobs,
                                                 inst.machines)
    solution_space = GanttSpace(inst)
    encoding = OperationBasedEncoding(inst)
    objective = Makespan(inst)
    if callable(algorithm):
        algorithm = algorithm(inst, search_space, objective)
    if not isinstance(algorithm, Algorithm):
        raise type_error(algorithm, "algorithm", Algorithm, call=True)

    goal: int
    if required_result is None:
        lb: int = objective.lower_bound()
        ub: int = objective.upper_bound()
        goal = max(lb + 1, min(ub - 1, int(0.5 + (lb + (0.96 * (ub - lb))))))
    else:
        goal = required_result

    validate_algorithm(algorithm=algorithm,
                       solution_space=solution_space,
                       objective=objective,
                       search_space=search_space,
                       encoding=encoding,
                       max_fes=max_fes,
                       required_result=goal,
                       post=post)


def validate_algorithm_on_jssp(
        algorithm: Callable[[Instance, Permutations,
                             Objective], Algorithm],
        max_fes: int = 100,
        post: Callable[[Algorithm, int], Any] | None = None) -> None:
    """
    Validate an algorithm on a set of JSSP instances.

    :param algorithm: the algorithm factory
    :param max_fes: the maximum FEs
    :param post: a check to run after each execution of the algorithm,
        receiving the algorithm and the number of consumed FEs as parameter
    """
    for i in jssp_instances_for_tests():
        validate_algorithm_on_1_jssp(algorithm, i, max_fes=max_fes, post=post)


def validate_objective_on_1_jssp(
        objective: Objective | Callable[[Instance], Objective],
        instance: str | None = None,
        is_deterministic: bool = True) -> None:
    """
    Validate an objective function on 1 JSSP instance.

    :param objective: the objective function or a factory creating it
    :param instance: the instance name
    :param is_deterministic: is the objective function deterministic?
    """
    if instance is None:
        instance = str(default_rng().choice(Instance.list_resources()))
    if not isinstance(instance, str):
        raise type_error(instance, "JSSP instance name", (str, None))
    inst = Instance.from_resource(instance)
    if not isinstance(inst, Instance):
        raise type_error(inst, f"loaded JSSP instance {instance!r}", Instance)

    if callable(objective):
        objective = objective(inst)

    validate_objective(
        objective=objective,
        solution_space=GanttSpace(inst),
        make_solution_space_element_valid=make_gantt_valid(inst),
        is_deterministic=is_deterministic)


def validate_objective_on_jssp(
        objective: Objective | Callable[[Instance], Objective],
        is_deterministic: bool = True) -> None:
    """
    Validate an objective function on JSSP instances.

    :param objective: the objective function or a factory creating it
    :param is_deterministic: is the objective function deterministic?
    """
    for i in jssp_instances_for_tests():
        validate_objective_on_1_jssp(objective, i, is_deterministic)


def validate_mo_algorithm_on_1_jssp(
        algorithm: MOAlgorithm | Callable[
            [Instance, Permutations, MOProblem], MOAlgorithm],
        instance: str | None = None, max_fes: int = 100) -> None:
    """
    Check the validity of a black-box multi-objective algorithm on the JSSP.

    :param algorithm: the algorithm or algorithm factory
    :param instance: the instance name, or `None` to randomly pick one
    :param max_fes: the maximum number of FEs
    """
    if not (isinstance(algorithm, MOAlgorithm) or callable(algorithm)):
        raise type_error(algorithm, "algorithm", MOAlgorithm, True)

    random: Final[Generator] = default_rng()
    if instance is None:
        instance = str(random.choice(Instance.list_resources()))
    if not isinstance(instance, str):
        raise type_error(instance, "JSSP instance name", (str, None))
    inst = Instance.from_resource(instance)
    if not isinstance(inst, Instance):
        raise type_error(inst, "loaded JSSP instance '{instance}'", Instance)

    search_space = Permutations.with_repetitions(inst.jobs,
                                                 inst.machines)
    solution_space = GanttSpace(inst)
    encoding = OperationBasedEncoding(inst)

    weights: Final[list[int | float]] = [float(random.uniform(0.01, 10)),
                                         float(random.uniform(0.01, 10))] \
        if random.integers(2) <= 0 else \
        [1 + int(random.integers(1 << random.integers(40))),
         1 + int(random.integers(1 << random.integers(40)))]
    problem: Final[MOProblem] = WeightedSum(
        [Makespan(inst), Worktime(inst)], weights)

    if callable(algorithm):
        algorithm = algorithm(inst, search_space, problem)
    if not isinstance(algorithm, MOAlgorithm):
        raise type_error(algorithm, "algorithm", MOAlgorithm, call=True)

    validate_mo_algorithm(algorithm=algorithm,
                          solution_space=solution_space,
                          problem=problem,
                          search_space=search_space,
                          encoding=encoding,
                          max_fes=max_fes)


def validate_mo_algorithm_on_jssp(
        algorithm: Callable[
            [Instance, Permutations, MOProblem], MOAlgorithm]) -> None:
    """
    Validate a multi-objective algorithm on a set of JSSP instances.

    :param algorithm: the algorithm factory
    """
    for i in jssp_instances_for_tests():
        validate_mo_algorithm_on_1_jssp(algorithm, i, 100)



"""
The EAFEA is hybrid of the (1+1)FEA and the (1+1)EA without Solution Transfer.

1. Thomas Weise, Zhize Wu, Xinlu Li, Yan Chen, and Jörg Lässig. Frequency
   Fitness Assignment: Optimization without Bias for Good Solutions can be
   Efficient. *IEEE Transactions on Evolutionary Computation (TEVC)*. 2022.
   Early Access. https://dx.doi.org/10.1109/TEVC.2022.3191698
"""
from collections import Counter
from typing import Callable, Final

from numpy.random import Generator
from pycommons.types import type_error

from moptipy.algorithms.so.ffa.ffa_h import create_h, log_h
from moptipy.api.algorithm import Algorithm1
from moptipy.api.operators import Op0, Op1
from moptipy.api.process import Process


class EAFEA(Algorithm1):
    """An implementation of the EAFEA."""

    def __init__(self, op0: Op0, op1: Op1, log_h_tbl: bool = False) -> None:
        """
        Create the EAFEA.

        :param op0: the nullary search operator
        :param op1: the unary search operator
        :param log_h_tbl: should we log the H table?
        """
        super().__init__("eafeaA", op0, op1)
        if not isinstance(log_h_tbl, bool):
            raise type_error(log_h_tbl, "log_h_tbl", bool)
        #: True if we should log the H table, False otherwise
        self.log_h_tbl: Final[bool] = log_h_tbl

    def solve(self, process: Process) -> None:
        """
        Apply the EAFEA to an optimization problem.

        :param process: the black-box process object
        """
        # Create records for old and new point in the search space.
        x_c = process.create()  # record for current solution of the EA
        x_d = process.create()  # record for current solution of the FEA
        x_n = process.create()  # record for new solution

        # Obtain the random number generator.
        random: Final[Generator] = process.get_random()

        # Put function references in variables to save time.
        evaluate: Final[Callable] = process.evaluate  # the objective
        should_terminate: Final[Callable] = process.should_terminate
        op0: Final[Callable] = self.op0.op0  # the nullary operator
        op1: Final[Callable] = self.op1.op1  # the unary operator

        h, ofs = create_h(process)  # Allocate the h-table

        # Start at a random point in the search space and evaluate it.
        op0(random, x_c)  # Create 1 solution randomly and
        y_c: int | float = evaluate(x_c) + ofs  # evaluate it.
        process.copy(x_d, x_c)
        y_d: int | float = y_c
        use_ffa: bool = True

        while not should_terminate():  # Until we need to quit...
            use_ffa = not use_ffa  # toggle use of FFA
            op1(random, x_n, x_d if use_ffa else x_c)
            y_n: int | float = evaluate(x_n) + ofs

            if use_ffa:  # the FEA branch
                h[y_n] += 1  # type: ignore  # Increase the frequency
                h[y_d] += 1  # type: ignore  # of new_f and cur_f.
                if h[y_n] <= h[y_d]:  # type: ignore
                    y_d = y_n  # Store its objective value.
                    x_d, x_n = x_n, x_d  # Swap best and new.
            elif y_n <= y_c:  # the EA/RLS branch
                y_c = y_n
                x_c, x_n = x_n, x_c

        if not self.log_h_tbl:
            return  # we are done here

        # After we are done, we want to print the H-table.
        if h[y_c] == 0:  # type: ignore  # Fix the H-table for the case
            h = Counter()   # that only one FE was performed: In this case,
            h[y_c] = 1  # make Counter with only a single 1 value inside.

        log_h(process, h, ofs)  # log the H-table



"""The setup and installation script."""

from re import compile as re_compile
from re import sub as re_sub
from typing import Final, Pattern

from setuptools import setup

# We want to use our README.md as project description.
# However, we must fix all the references inside.
with open("README.md", encoding="utf-8-sig") as reader:
    old_lines = reader.readlines()

# It seems that the markdown parser does not auto-generate anchors. This means
# that we need to fix all references following the pattern `[xxx](#12-hello)`
# to `[xxx]({docu_url#hello)`, where `docu_url` is the url of our
# documentation. We do this with a regular expression `regex_search`.
new_lines: Final[list[str]] = []
in_code: bool = False  # we only process non-code lines

# the base url where the documentation will land
doc_url: Final[str] = "https://thomasweise.github.io/moptipy"
# the url of our repository
repo_url: Final[str] = "https://github.com/thomasWeise/moptipy"

# detects strings of the form [xyz](#123-bla) and gives \1=xyz and \2=bla
regex_search: Final[Pattern] = re_compile("(\\[.+?])\\(#\\d+-(.+?)\\)")
regex_repl: Final[str] = f"\\1({doc_url}#\\2)"

# other replacements
license_old: Final[str] = f"{repo_url}/blob/main/LICENSE"
license_new: Final[str] = f"{doc_url}/LICENSE.html"

for full_line in old_lines:
    line: str = str.rstrip(full_line)
    if in_code:
        if line.startswith("```"):
            in_code = False  # toggle to non-code
    elif line.startswith("```"):
        in_code = True  # toggle to code
    else:  # fix all internal urls
        # replace links of the form "#12-bla" to "#bla"
        line = re_sub(regex_search, regex_repl, line)
        line = str.replace(line, license_old, license_new)
    new_lines.append(line)

# Now we can use the code in the setup.
setup(long_description="\n".join(new_lines),
      long_description_content_type="text/markdown")



"""A simple record for storing a solution with its quality."""

from typing import Final


# start record
class Record:
    """
    A point in the search space, its quality, and creation time.

    A record stores a point in the search space :attr:`x` together with
    the corresponding objective value :attr:`f`. It also stores a
    "iteration index" :attr:`it`, i.e., the time when the point was
    created or modified.

    This allows for representing and storing solutions in a population.
    If the population is sorted, then records with better objective
    value will be moved to the beginning of the list. Ties are broken
    such that younger individuals (with higher :attr:`it` value) are
    preferred.
    """

    def __init__(self, x, f: int | float):
        """
        Create the record.

        :param x: the data structure for a point in the search space
        :param f: the corresponding objective value
        """
        #: the point in the search space
        self.x: Final = x
        #: the objective value corresponding to :attr:`x`
        self.f: int | float = f
        #: the iteration index when :attr:`x` was sampled
        self.it: int = 0

    def __lt__(self, other) -> bool:
        """
        Precedence if 1) better or b) equally good but younger.

        :param other: the other record
        :returns: `True` if this record has a better objective value
            (:attr:`f`) or if it has the same objective value but is newer,
            i.e., has a larger :attr:`it` value

        >>> r1 = Record(None, 10)
        >>> r2 = Record(None, 9)
        >>> r1 < r2
        False
        >>> r2 < r1
        True
        >>> r1.it = 22
        >>> r2.f = r1.f
        >>> r2.it = r1.it
        >>> r1 < r2
        False
        >>> r2 < r1
        False
        >>> r2.it = r1.it + 1
        >>> r1 < r2
        False
        >>> r2 < r1
        True
        """
        f1: Final[int | float] = self.f
        f2: Final[int | float] = other.f
        return (f1 < f2) or ((f1 == f2) and (self.it > other.it))
# end record



"""
Provides the Algorithm `bobyqa` from the Library `pdfo`.

The library "Powell's Derivative-Free Optimization solvers" (`pdfo`) at
https://www.pdfo.net provides an implementation of the "Bound Optimization BY
Quadratic Approximation" algorithm, or BOBYQA for short.
The library is dedicated to the late Professor M. J. D. Powell FRS (1936—2015)
and maintained by Tom M. Ragonneau and Zaikun Zhang.
Here, we wrap it into a class that complies to our `moptipy` API.
This class offers no additional logic but directly defers to the function
in the `pdfo` library.

1. Michael James David Powell. The BOBYQA Algorithm for Bound Constrained
   Optimization without Derivatives. Technical Report DAMTP 2009/NA06.
   Department of Applied Mathematics and Theoretical Physics, Cambridge
   University, Cambridge, UK, 2009.
   https://www.damtp.cam.ac.uk/user/na/NA_papers/NA2009_06.pdf
2. Tom M. Ragonneau and Zaikun Zhang. *PDFO: a cross-platform package for
   Powell's derivative-free optimization solvers,* arXiv preprint.
   Ithaca, NY, USA: Cornell University Library February, 2023.
   arXiv:2302.13246v1 [math.OC] 26 Feb 202.
   https://arxiv.org/pdf/2302.13246v1

- https://github.com/pdfo/pdfo
- https://www.pdfo.net
- https://pypi.org/project/pdfo
"""

import warnings
from typing import Any, Callable, Final, cast

import numpy as np
import pdfo  # type: ignore

# noinspection PyProtectedMember
from pdfo._bobyqa import bobyqa  # type: ignore  # noqa: PLC2701
from pycommons.types import type_error

from moptipy.api.algorithm import Algorithm0
from moptipy.api.operators import Op0
from moptipy.api.process import Process
from moptipy.api.subprocesses import (
    get_remaining_fes,
    without_should_terminate,
)
from moptipy.spaces.vectorspace import VectorSpace
from moptipy.utils.logger import KeyValueLogSection

#: pdfo with version 1.3 and below is incompatible with numpy
#: of version 1.24.0 and above. It will crash with an exception.
#: So for this case, we will later just invoke a single random sample and
#: exit. See https://github.com/pdfo/pdfo/issues/55
_CANNOT_DO_PDFO: Final[bool] = \
    hasattr(np, "__version__") and hasattr(pdfo, "__version__") \
    and (list(map(int, np.__version__.split("."))) >= [1, 24])\
    and (list(map(int, pdfo.__version__.split("."))) <= [1, 3])


class BOBYQA(Algorithm0):
    """
    A wrapper for the `bobyqa` algorithm from `pdfo`.

    The Bound Optimization BY Quadratic Approximation (BOBYQA) developed by
    Michael James David Powell and published by the `pdfo` library.

    1. Michael James David Powell. The BOBYQA Algorithm for Bound Constrained
       Optimization without Derivatives. Technical Report DAMTP 2009/NA06.
       Department of Applied Mathematics and Theoretical Physics, Cambridge
       University, Cambridge, UK, 2009.
       https://www.damtp.cam.ac.uk/user/na/NA_papers/NA2009_06.pdf
    """

    def __init__(self, op0: Op0, space: VectorSpace) -> None:
        """
        Create the BOBYQA algorithm.

        :param op0: the nullary search operator
        :param space: the vector space
        """
        super().__init__("bobyqa_pdfo", op0)
        if not isinstance(space, VectorSpace):
            raise type_error(space, "space", VectorSpace)
        #: the vector space defining the dimensions and bounds
        self.space: Final[VectorSpace] = space

    def __run(self, process: Process) -> None:
        """
        Execute the algorithm.

        :param process: the process
        """
        x0: Final[np.ndarray] = process.create()
        npt: int = (2 * len(x0)) + 1  # the default npt value
        max_fes: int = max(npt + 1, get_remaining_fes(process))
        self.op0.op0(process.get_random(), x0)  # sample start point

        if _CANNOT_DO_PDFO:  # PDFO incompatible to current setup
            process.evaluate(x0)  # do single random sample
            return  # and quit

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            bobyqa(self.space.clipped(process.evaluate),
                   x0, bounds=np.stack((self.space.lower_bound,
                                        self.space.upper_bound)).transpose(),
                   options={
                       "rhoend": 1e-320,  # the end rho value
                       "maxfev": max_fes,  # the maximum FEs
                       "npt": npt,  # the number of interpolation points
                       "honour_x0": True,  # enforce our x0
                       "quiet": True})  # do not print any messages

    def solve(self, process: Process) -> None:
        """
        Apply the external `bobyqa` implementation to an optimization problem.

        :param process: the black-box process object
        """
        # invoke the SciPy algorithm implementation
        without_should_terminate(
            cast(Callable[[Process], Any], self.__run), process)

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of the algorithm to a logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)  # log algorithm/operator
        self.space.log_bounds(logger)  # log bounds



"""The well-known Trap problem."""
from typing import Final

import numba  # type: ignore
import numpy as np

from moptipy.examples.bitstrings.bitstring_problem import BitStringProblem


@numba.njit(nogil=True, cache=True)
def trap(x: np.ndarray) -> int:
    """
    Compute the trap objective value.

    :param x: the np array
    :return: the trap function value

    >>> print(trap(np.array([True, True, False, False, False])))
    3
    >>> print(trap(np.array([True, False, True, False, False])))
    3
    >>> print(trap(np.array([False, True,  True, False, False])))
    3
    >>> print(trap(np.array([True, True, True, True, True])))
    0
    >>> print(trap(np.array([False, True, True, True, True])))
    5
    >>> print(trap(np.array([False, False, False, False, False])))
    1
    >>> print(trap(np.array([False, True,  False, False, False])))
    2
    >>> print(trap(np.array([False, True,  True, True, False])))
    4
    """
    length: Final[int] = len(x)
    s: Final[int] = x.sum()
    return 0 if (s >= length) else int(s + 1)


class Trap(BitStringProblem):
    """The trap problem."""

    def __init__(self, n: int) -> None:  # +book
        """
        Initialize the trap objective function.

        :param n: the dimension of the problem

        >>> print(Trap(55).n)
        55
        >>> print(Trap(4).evaluate(np.array([True, True, False, True])))
        4
        """
        super().__init__(n)
        self.evaluate = trap  # type: ignore

    def __str__(self) -> str:
        """
        Get the name of the trap objective function.

        :return: `trap_` + length of string

        >>> print(Trap(33))
        trap_33
        """
        return f"trap_{self.n}"



"""An implementation of an box-constrained n-dimensional continuous space."""

from math import isfinite
from typing import Callable, Final, Iterable, cast

import numpy as np
from numpy import clip
from pycommons.types import type_error

from moptipy.spaces.nparrayspace import NPArraySpace
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.nputils import DEFAULT_FLOAT, array_to_str, is_np_float

#: the log key for the lower bound, i.e., the minimum permitted value
KEY_LOWER_BOUND: Final[str] = "lb"
#: the log key for the upper bound, i.e., the maximum permitted value
KEY_UPPER_BOUND: Final[str] = "ub"


class VectorSpace(NPArraySpace):
    """
    A vector space where each element is an n-dimensional real vector.

    Such spaces are useful for continuous optimization. The vectors are
    implemented as one-dimensional :class:`numpy.ndarray` of length `n`.
    A vector space is constraint by a box which defines the minimum and
    maximum permitted value for each of its `n` elements.

    >>> s = VectorSpace(3)
    >>> print(s.dimension)
    3
    >>> print(s.dtype)
    float64
    >>> print(s.lower_bound)
    [0. 0. 0.]
    >>> print(s.upper_bound)
    [1. 1. 1.]
    >>> print(s.lower_bound_all_same)
    True
    >>> print(s.upper_bound_all_same)
    True
    >>> s = VectorSpace(2, -1.0, 5.0)
    >>> print(s.lower_bound)
    [-1. -1.]
    >>> print(s.upper_bound)
    [5. 5.]
    >>> s = VectorSpace(2, [-1.0, -2.0], 5.0)
    >>> print(s.lower_bound)
    [-1. -2.]
    >>> print(s.upper_bound)
    [5. 5.]
    >>> print(s.lower_bound_all_same)
    False
    >>> print(s.upper_bound_all_same)
    True
    """

    def __init__(self, dimension: int,
                 lower_bound: float | Iterable[float] = 0.0,
                 upper_bound: float | Iterable[float] = 1.0,
                 dtype: np.dtype = DEFAULT_FLOAT) -> None:
        """
        Create the vector-based search space.

        :param dimension: The dimension of the search space,
            i.e., the number of decision variables.
        :param dtype: The basic data type of the vector space,
            i.e., the type of the decision variables
        :param lower_bound: the optional minimum value(s)
        :param upper_bound: the optional maximum value(s)
        """
        super().__init__(dimension, dtype)
        if not is_np_float(dtype):
            raise TypeError(f"Invalid data type {dtype}.")

        # first, we process the lower bounds
        lower_bound_all_same: bool = True
        if isinstance(lower_bound, float):
            # only a single value is given
            if not isfinite(lower_bound):
                raise ValueError(
                    f"invalid lower bound {lower_bound}.")
            # if a single value is given, then we expand it a vector
            lower_bound = np.full(dimension, lower_bound, dtype)
        elif isinstance(lower_bound, Iterable):
            # lower bounds are given as vector or iterable
            lb = np.array(lower_bound, dtype)
            if len(lb) != dimension:
                raise ValueError(f"wrong length {lb} of lower "
                                 f"bound iterable {lower_bound}")
            if lb.shape != (dimension, ):
                raise ValueError(f"invalid shape={lb.shape} of "
                                 f"lower bound {lower_bound}")
            first = lb[0]
            for index, item in enumerate(lb):
                if first != item:
                    lower_bound_all_same = False
                if not np.isfinite(item):
                    raise ValueError(f"{index}th lower bound={item}")
            lower_bound = lb
        else:
            raise type_error(lower_bound, "lower_bound", (
                float, Iterable, None))

        # now, we process the upper bounds
        upper_bound_all_same: bool = True
        if isinstance(upper_bound, float):
            # only a single value is given
            if not isfinite(upper_bound):
                raise ValueError(
                    f"invalid upper bound {upper_bound}.")
            # if a single value is given, then we expand it a vector
            upper_bound = np.full(dimension, upper_bound, dtype)
        elif isinstance(upper_bound, Iterable):
            # upper bounds are given as vector or iterable
            lb = np.array(upper_bound, dtype)
            if len(lb) != dimension:
                raise ValueError(f"wrong length {lb} of upper "
                                 f"bound iterable {upper_bound}")
            if lb.shape != (dimension,):
                raise ValueError(f"invalid shape={lb.shape} of "
                                 f"upper bound {upper_bound}")
            first = lb[0]
            for index, item in enumerate(lb):
                if first != item:
                    upper_bound_all_same = False
                if not np.isfinite(item):
                    raise ValueError(f"{index}th upper bound={item}")
            upper_bound = lb
        else:
            raise type_error(upper_bound, "upper_bound", (
                float, Iterable, None))

        # check that the bounds are consistent
        for idx, ll in enumerate(lower_bound):
            if not ll < upper_bound[idx]:
                raise ValueError(f"lower_bound[{idx}]={ll} >= "
                                 f"upper_bound[{idx}]={upper_bound[idx]}")

        #: the lower bounds for all variables
        self.lower_bound: Final[np.ndarray] = lower_bound
        #: all dimensions have the same lower bound
        self.lower_bound_all_same: Final[bool] = lower_bound_all_same
        #: the upper bounds for all variables
        self.upper_bound: Final[np.ndarray] = upper_bound
        #: all dimensions have the same upper bound
        self.upper_bound_all_same: Final[bool] = upper_bound_all_same

    def clipped(self, func: Callable[[np.ndarray], int | float]) \
            -> Callable[[np.ndarray], int | float]:
        """
        Wrap a function ensuring that all vectors are clipped to the bounds.

        This function is useful to ensure that only valid vectors are passed
        to :meth:`~moptipy.api.process.Process.evaluate`.

        :param func: the function to wrap
        :returns: the wrapped function
        """
        return cast(Callable[[np.ndarray], int | float],
                    lambda x, lb=self.lower_bound, ub=self.upper_bound,
                    ff=func: ff(clip(x, lb, ub, x)))

    def validate(self, x: np.ndarray) -> None:
        """
        Validate a vector.

        :param x: the real vector
        :raises TypeError: if the string is not an :class:`numpy.ndarray`.
        :raises ValueError: if the shape of the vector is wrong or any of its
            element is not finite.
        """
        super().validate(x)

        mib: Final[np.ndarray] = self.lower_bound
        mab: Final[np.ndarray] = self.upper_bound
        for index, item in enumerate(x):
            miv = mib[index]
            mav = mab[index]
            if not np.isfinite(item):
                raise ValueError(f"x[{index}]={item}, which is not finite")
            if not (miv <= item <= mav):
                raise ValueError(
                    f"x[{index}]={item}, but should be in [{miv},{mav}].")

    def n_points(self) -> int:
        """
        Get an upper bound for the number of different values in this space.

        :return: We return the approximate number of finite floating point
          numbers while ignoring the box constraint. This value here therefore
          is an upper bound.

        >>> import numpy as npx
        >>> print(VectorSpace(3, dtype=npx.dtype(npx.float64)).n_points())
        6267911251143764491534102180507836301813760039183993274367
        """
        if self.dtype.char == "e":
            exponent = 5
            mantissa = 10
        elif self.dtype.char == "f":
            exponent = 8
            mantissa = 23
        elif self.dtype == "d":
            exponent = 11
            mantissa = 52
        elif self.dtype == "g":
            exponent = 15
            mantissa = 112
        else:
            raise ValueError(f"Invalid dtype {self.dtype}.")

        base = 2 * ((2 ** exponent) - 1) * (2 ** mantissa) - 1
        return base ** self.dimension

    def __str__(self) -> str:
        """
        Get the name of this space.

        :return: "r" + dimension + dtype.char

        >>> import numpy as npx
        >>> print(VectorSpace(3, dtype=npx.dtype(npx.float64)))
        r3d
        """
        return f"r{self.dimension}{self.dtype.char}"

    def log_bounds(self, logger: KeyValueLogSection) -> None:
        """
        Log the bounds of this space to the given logger.

        :param logger: the logger for the parameters

        >>> from moptipy.utils.logger import InMemoryLogger
        >>> import numpy as npx
        >>> space = VectorSpace(2, -5.0, [2.0, 3.0])
        >>> with InMemoryLogger() as l:
        ...     with l.key_values("C") as kv:
        ...         space.log_bounds(kv)
        ...     text = l.get_log()
        >>> text[-2]
        'ub: 2;3'
        >>> text[-3]
        'lb: -5'
        >>> len(text)
        4
        """
        if self.lower_bound_all_same:
            logger.key_value(
                KEY_LOWER_BOUND, self.lower_bound[0], also_hex=False)
        else:
            logger.key_value(KEY_LOWER_BOUND, array_to_str(self.lower_bound))
        if self.upper_bound_all_same:
            logger.key_value(
                KEY_UPPER_BOUND, self.upper_bound[0], also_hex=False)
        else:
            logger.key_value(KEY_UPPER_BOUND, array_to_str(self.upper_bound))

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of this space to the given logger.

        :param logger: the logger for the parameters

        >>> from moptipy.utils.logger import InMemoryLogger
        >>> import numpy as npx
        >>> space = VectorSpace(2, -5.0, [2.0, 3.0])
        >>> space.dimension
        2
        >>> space.dtype.char
        'd'
        >>> with InMemoryLogger() as l:
        ...     with l.key_values("C") as kv:
        ...         space.log_parameters_to(kv)
        ...     text = l.get_log()
        >>> text[-2]
        'ub: 2;3'
        >>> text[-3]
        'lb: -5'
        >>> text[-4]
        'dtype: d'
        >>> text[-5]
        'nvars: 2'
        >>> text[-6]
        'class: moptipy.spaces.vectorspace.VectorSpace'
        >>> text[-7]
        'name: r2d'
        >>> len(text)
        8
        """
        super().log_parameters_to(logger)
        self.log_bounds(logger)
